from __future__ import annotations

from ..scenes.atmosphere import (
    AbstractHeterogeneousAtmosphere,
    Atmosphere,
    MolecularAtmosphere,
)
from ..scenes.bsdfs import BSDF, bsdf_factory
from ..scenes.geometry import SceneGeometry, SphericalShellGeometry
from ..scenes.measure import (
    AbstractDistantMeasure,
    Measure,
    MultiRadiancemeterMeasure,
    RadiancemeterMeasure,
)
from ..scenes.shapes import RectangleShape
from ..scenes.surface import BasicSurface, Surface, surface_factory


def measure_inside_atmosphere(atmosphere: Atmosphere, measure: Measure) -> bool:
    """
    Evaluate whether a sensor is placed within an atmosphere.

    Raises a ValueError if called with a :class:`.MultiRadiancemeterMeasure`
    with origins both inside and outside the atmosphere.
    """
    if atmosphere is None:
        return False

    shape = atmosphere.shape

    if isinstance(measure, MultiRadiancemeterMeasure):
        inside = shape.contains(measure.origins)

        if all(inside):
            return True

        elif not any(inside):
            return False

        else:
            raise ValueError(
                "Inconsistent placement of MultiRadiancemeterMeasure origins. "
                "Origins must lie either all inside or all outside of the "
                "atmosphere."
            )

    elif isinstance(measure, AbstractDistantMeasure):
        # Note: This will break if the user makes something weird such as using
        # a large offset value which would put some origins outside and others
        # inside the atmosphere shape
        return not measure.is_distant()

    elif isinstance(measure, RadiancemeterMeasure):
        return shape.contains(measure.origin)

    else:
        # Note: This will likely break if a new measure type is added
        return shape.contains(measure.origin)


def surface_converter(value: dict | Surface | BSDF) -> Surface:
    """
    Attempt to convert the surface specification into a surface type.

    Surfaces can be defined purely by their BSDF, in which case Eradiate will
    define a rectangular surface and attach that BSDF to it.
    """
    if isinstance(value, dict):
        try:
            # First, attempt conversion to BSDF
            value = bsdf_factory.convert(value)
        except ValueError:  # Type ID could not be found in BSDF factory registry
            # Attempt conversion to Surface
            return surface_factory.convert(value)

    # If we make it to this point, it means that dict conversion has been
    # performed with success
    if isinstance(value, BSDF):
        return BasicSurface(
            shape=RectangleShape(),
            bsdf=value,
        )

    return value


def check_geometry_atmosphere(
    geometry: SceneGeometry, atmosphere: MolecularAtmosphere
) -> None:
    """
    Check that the experiment geometry is compatible with the vertical extent of
    the molecular atmosphere.

    Parameters
    ----------
    geometry : SceneGeometry
        An experiment geometry.

    atmosphere : MolecularAtmosphere
        A molecular atmosphere.

    Raises
    ------
    ValueError
        If the geometry vertical extent exceeds the atmosphere vertical
        extent.
    """

    radprops_zbounds = atmosphere.radprops_profile.zbounds
    geometry_zbounds = geometry.zgrid.levels[[0, -1]]
    suggested_solution = (
        "Try to set the experiment geometry so that it does not go beyond "
        "the vertical extent of the molecular atmosphere."
    )
    if (geometry_zbounds[0] < radprops_zbounds[0]) or (
        geometry_zbounds[1] > radprops_zbounds[1]
    ):
        raise ValueError(
            "Attribtues 'geometry' and 'atmosphere' are incompatible: "
            f"'geometry.zgrid' bounds ({geometry_zbounds}) go beyond the "
            f"bounds of 'atmosphere.radprops_profile.zbounds' ({radprops_zbounds}). "
            f"{suggested_solution}"
        )


def check_piecewise_compatible(
    geometry: SceneGeometry, atmosphere: Atmosphere | None
) -> tuple[bool, str]:
    """
    Check that the piecewise integrator is compatible with the experiment's
    atmosphere and geometry.

    Parameters
    ----------
    geometry : SceneGeometry
        An experiment geometry.

    atmosphere : Atmosphere
        The scene's atmosphere configuration.

    Returns
    -------
    tuple[bool, str]
        Piecewise compatible flag and an exception message if the flag is not raised.
        The message is empty if the flag is raised.
    """

    if isinstance(geometry, SphericalShellGeometry):
        debug_message = (
            "Piecewise integrator not compatible with spherical shell geometry."
        )
        return False, debug_message

    if atmosphere is None or not isinstance(
        atmosphere, AbstractHeterogeneousAtmosphere
    ):
        debug_message = f"Piecewise integrator not compatible with medium type : {type(atmosphere)}."
        return False, debug_message

    if atmosphere.force_majorant:
        debug_message = "Piecewise integrator not compatible with a medium that has force_majorant = True."
        return False, debug_message

    return True, ""


def check_path_compatible(atmosphere: Atmosphere | None) -> tuple[bool, str]:
    """
    Check if the scene is compatible with the ``path`` integrator.

    Parameters
    ----------
    atmosphere : Atmosphere
        The scene's atmosphere configuration.

    Returns
    -------
    tuple[bool, str]
        Path compatible flag and an exception message if the flag is not raised.
        The message is empty if the flag is raised.
    """

    if atmosphere is not None:
        debug_message = (
            "You have an atmosphere in your scene, prefer using the Path Integrator."
        )
        return False, debug_message

    return True, ""



"""
Helper tools for plugin tests.
"""

from __future__ import annotations

import drjit as dr
import mitsuba as mi


def sample_eval_pdf_bsdf(
    plugin: mi.BSDF,
    wi: mi.ScalarVector3f,
    sample_count: int = 100000,
    seed: int = 0,
):
    """
    Sample a BSDF and get corresponding ``eval()`` and ``pdf()`` values.

    Parameters
    ----------
    plugin : mitsuba.BSDF
        A BSDF plugin to be evaluated.

    wi : mitsuba.ScalarVector3f
        An incoming direction.

    sample_count : int, optional, default: 100000
        The number of samples to generate.

    seed : int, optional, default: 0
        RNG seed to be used.

    Returns
    -------
    sample
        The result of a call to the plugin's ``sample()`` method.

    eval
        The result of a call to the plugin's ``eval()`` method.

    pdf
        The result of a call to the plugin's ``pdf()`` method.

    Warnings
    --------
    This function requires a JIT active variant (*e.g.* ``llvm_ad_rgb``).
    """

    # Generate a table of uniform variates
    idx = dr.arange(mi.UInt64, sample_count)
    v0, v1 = mi.sample_tea_32(idx, seed)

    # Scramble seed and stream index using TEA, a linearly increasing sequence
    # does not produce a sufficiently statistically independent set of streams
    rng = mi.PCG32(initstate=v0, initseq=v1)

    sample = mi.Vector3f()
    for i in range(3):
        sample[i] = rng.next_float32() if mi.Float is mi.Float32 else rng.next_float64()

    # Invoke sampling strategy
    n = dr.width(sample)
    ctx = mi.BSDFContext()
    si = dr.zeros(mi.SurfaceInteraction3f, n)
    si.wi = wi
    bs, weight = plugin.sample(ctx, si, sample[0], [sample[1], sample[2]])

    # Get corresponding eval() and pdf() calls
    eval = plugin.eval(ctx, si, bs.wo)
    pdf = plugin.pdf(ctx, si, bs.wo)

    return (bs, weight), eval, pdf



"""Functions to compute Rayleigh scattering in air."""

import numpy as np
import numpy.typing as npt
import pint
from scipy.constants import physical_constants

from .. import converters
from ..units import unit_registry as ureg
from ..util.misc import Singleton

# Physical constants
#: Loschmidt constant [km^-3].
_LOSCHMIDT = ureg.Quantity(
    *physical_constants["Loschmidt constant (273.15 K, 101.325 kPa)"][:2]
).to("km^-3")

# Air number density at 101325 Pa and 288.15 K
_STANDARD_AIR_NUMBER_DENSITY = _LOSCHMIDT * (273.15 / 288.15)


# Bates (1984) King correction factor data
class _BATES_1984_DATA(metaclass=Singleton):
    # Lazy loader for Bates (1984) King correction data
    def __init__(self):
        self._king_factor = None
        self._wavelength = None

    @property
    def king_factor(self):
        if self._king_factor is None:
            self._load_dataset()
        return self._king_factor

    @property
    def wavelength(self):
        if self._wavelength is None:
            self._load_dataset()
        return self._wavelength

    def interp(self, w: npt.ArrayLike = np.array([0.550])):
        """
        Interpolate ``king_factor`` over the ``wavelength`` dimension.

        Parameters
        ----------
        w : array-like
            wavelengths (in micron) at wich to interpolate Bates' King factor.

        Returns
        -------
        ndarray
            Bates' King factor
        """
        left = self.king_factor[0]
        right = self.king_factor[-1]

        return np.interp(w, self.wavelength, self.king_factor, left=left, right=right)

    def _load_dataset(self):
        bates_data = converters.load_dataset("constant/optics/bates_1984.nc")
        self._king_factor = bates_data.f.values
        self._wavelength = bates_data.w.values


def compute_sigma_s_air(
    wavelength: pint.Quantity = ureg.Quantity(550.0, "nm"),
    number_density: pint.Quantity = _STANDARD_AIR_NUMBER_DENSITY,
) -> pint.Quantity:
    r"""
    Compute the Rayleigh scattering coefficient of air.

    When default values are used, this provides the Rayleigh scattering
    coefficient for air at 550 nm in standard temperature and pressure
    conditions.

    The scattering coefficient is computed by considering the air as a pure
    gas with associated effective optical properties (refractive index,
    King factor) and according to the expression provided by
    :cite:`Eberhard2010CorrectEquationsCommon` (eq. 60):

    .. math::

       k_{\mathrm s \, \lambda} (n) = \frac{8 \pi^3}{3 \lambda^4} \frac{1}{n}
          \left( \eta_{\lambda}^2(n) - 1 \right)^2 F_{\lambda}

    where
    :math:`\lambda` is the wavelength (subscript indicates spectral dependence),
    :math:`n` is the air number density,
    :math:`\eta` is the air refractive index and
    :math:`F` is the air King factor.

    The King correction factor is computed by linearly interpolating the data
    from :cite:`Bates1984RayleighScatteringAir`.

    Parameters
    ----------
    wavelength : quantity
        Wavelength [nm].

    number_density : quantity
        Number density of the scattering particles [km^-3].

    Returns
    -------
    quantity
        Scattering coefficient.
    """
    # We are going to elevate `wavelength` to the power 4: if it is stored as a
    # 32-bit int, as can happen on Windows where the default integer type is
    # int32, calculus will be wrong due to integer overflow. To mitigate that
    # risk, we convert the wavelength to micron.
    # In addition, the Bates (1984) dataset is indexed by wavelengths in microns
    # as well, meaning that this conversion is anyway necessary.
    w = wavelength.to("micron")

    BATES_1984_DATA = _BATES_1984_DATA()
    king_factor = BATES_1984_DATA.interp(w.m_as("micron"))

    refractive_index = air_refractive_index(wavelength=w, number_density=number_density)
    if isinstance(w.magnitude, np.ndarray) and isinstance(
        number_density.magnitude, np.ndarray
    ):
        king_factor = king_factor[:, np.newaxis]
        w = w[:, np.newaxis]
        number_density = number_density[np.newaxis, :]

    result = (
        8.0
        * np.power(np.pi, 3)
        / (3.0 * np.power(w, 4))
        / number_density
        * np.square(np.square(refractive_index) - 1.0)
        * king_factor
    )
    return result.to("km^-1")


def air_refractive_index(
    wavelength: pint.Quantity = ureg.Quantity(550.0, "nm"),
    number_density: pint.Quantity = _STANDARD_AIR_NUMBER_DENSITY,
) -> np.ndarray:
    """
    Computes the air refractive index.

    The wavelength dependence of the refractive index is computed using equation
    2 from :cite:`Peck1972DispersionAir`. This formula is a fit of
    measurements of the air refractive index in the range of wavelength from
    :math:`\\lambda = 240` nm to :math:`1690` nm.
    The number density dependence is computed using a simple proportionality
    rule.

    Parameters
    ----------
    wavelength : quantity
        Wavelength.

    number_density : quantity
        Number density.

        Default: Air number density at 101325 Pa and 288.15 K.

    Returns
    -------
    float or ndarray
        Air refractive index value(s).
    """

    # wavenumber in inverse micrometer
    sigma = 1 / wavelength.m_as("micrometer")
    sigma2 = np.square(sigma)

    # refractivity in parts per 1e8
    x = (5791817.0 / (238.0183 - sigma2)) + 167909.0 / (57.362 - sigma2)

    if isinstance(x, np.ndarray) and isinstance(number_density.magnitude, np.ndarray):
        x = x[:, np.newaxis]
        number_density = number_density[np.newaxis, :]

    # number density scaling
    x_scaled = x * (number_density / _STANDARD_AIR_NUMBER_DENSITY).m_as("dimensionless")

    # refractive index
    index = 1 + x_scaled * 1e-8

    return index


def depolarization_bates(wavelength: pint.Quantity = ureg.Quantity(550.0, "nm")):
    """
    Compute depolarization using Bates' King factor :cite:`Bates1984RayleighScatteringAir`.
    Only parametrized on wavelength.

    Parameters
    ----------
    wavelength : quantity
        Wavelength.

    Returns
    -------
    Quantity : scalar
        The depolarization factor parametrized on the wavelength. [dimensionless]
    """
    # The Bates (1984) dataset is indexed by wavelengths in microns
    # as well, meaning that this conversion is anyway necessary.
    w = wavelength.m_as("micron")

    BATES_1984_DATA = _BATES_1984_DATA()
    king_factor = BATES_1984_DATA.interp(w)

    depol = 6 * (king_factor - 1) / (7 * king_factor + 3)
    return np.atleast_1d(depol) * ureg.dimensionless


def depolarization_bodhaine(
    wavelength: pint.Quantity = ureg.Quantity(550.0, "nm"),
    x_CO2: pint.Quantity = ureg.Quantity(0.0004, "dimensionless"),
):
    """
    Compute depolarization using Bodhaine's King factor :cite:p:`Bodhaine1999RayleighOpticalDepth`.
    Parametrized over wavelength and CO2 concentration, other components
    are assumed to be at a fixed proportion. Valid at 273.15 K,
    1013.25 mb, for dry air.

    Parameters
    ----------
    wavelength : quantity
        Wavelength.

    x_CO2 : quantity
        Array of CO2 mole fraction in the atmosphere.

    Returns
    -------
    quantity : array of shape (N,)
        Array of depolarization factor for each level [dimensionless]
    """
    w_um = wavelength.m_as("um")

    # part per volume by percent
    C_CO2 = x_CO2.m_as("%")
    total = 78.084 + 20.946 + 0.934 + C_CO2

    # from Bates (1984), wavelength is in micron
    F_N2 = 1.034 + 3.17 * 1e-4 * 1 / w_um**2
    F_O2 = 1.096 + 1.385 * 1e-3 * 1 / w_um**2 + 1.448 * 1e-4 * 1 / w_um**4
    F_air = (78.084 * F_N2 + 20.946 * F_O2 + 0.934 * 1.00 + C_CO2 * 1.15) / total

    # calculate the depolarization
    return 6 * (F_air - 1) / (7 * F_air + 3) * ureg.dimensionless



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from io import StringIO
from pathlib import Path
from typing import ClassVar

import attrs
import matplotlib.pyplot as plt
import mitsuba as mi
import numpy as np
import scipy.stats as spstats
import xarray as xr
from numpy.typing import ArrayLike
from robot.api import logger

from .. import fresolver
from ..attrs import define, documented
from ..exceptions import DataError
from ..typing import PathLike
from ..util.misc import summary_repr


def regression_test_plots(
    ref: ArrayLike,
    result: ArrayLike,
    vza: ArrayLike,
    metric: tuple[str, float],
    ref_var: ArrayLike | None = None,
    result_var: ArrayLike | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
) -> tuple[plt.Figure, list[list[plt.Axes]]]:
    """
    Create regression test report plots. Plot errorbars if both ref_var and
    result_var are set.

    Parameters
    ----------
    ref : array-like
        Variable values for the reference data

    result : array-like
        Variable values for the simulation result

    vza : array-like
        VZA values for plotting

    metric : tuple
        A tuple of the form (metric name, value) to be added to the plots.

    ref_var : array-like, optional
        Variable variance for the reference data.

    result_var : array-like, optional
        Variable variance for the simulation result.

    xlabel, ylabel : str or None
        Labels applied to the x and y axes of the plot.

    Returns
    -------
    figure: Figure
        Pyplot Figure containing the report charts

    axes: list
        2x2 array of Axes included in the report Figure
    """
    fig, axes = plt.subplots(2, 2, figsize=(8, 6), layout="constrained")

    ax = axes[0][0]
    if ref_var is None:
        ax.plot(vza, ref, label="reference")
    else:
        ax.errorbar(vza, ref, yerr=np.sqrt(ref_var), label="reference")

    if result_var is None:
        ax.plot(vza, result, label="result")
    else:
        ax.errorbar(vza, result, yerr=np.sqrt(result_var), label="result")

    ax.set_title("Reference and test result")
    handles, labels = ax.get_legend_handles_labels()

    ax = axes[1][0]
    ax.plot(vza, result - ref)
    ax.set_title("Absolute difference")

    ax = axes[1][1]
    ax.plot(vza, (result - ref) / ref)
    ax.set_title("Relative difference")

    ax = axes[0][1]
    ax.set_axis_off()
    ax.legend(handles=handles, labels=labels, loc="upper center")

    if metric[1] is None:
        ax.text(
            0.5,
            0.5,
            f'Metric "{metric[0]}" is not available',
            horizontalalignment="center",
        )
    else:
        ax.text(0.5, 0.5, f"{metric[0]}: {metric[1]:.4}", horizontalalignment="center")

    for i, j in [[0, 0], [1, 0], [1, 1]]:
        ax = axes[i][j]
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        if ylabel is not None:
            ax.set_ylabel(ylabel)

    return fig, axes


def figure_to_html(fig: plt.Figure) -> str:
    """
    Render a figure in HTML format

    Returns a string containing the rendered HTML. The root tag is a <svg> one.

    Parameters
    ----------
    fig : plt.Figure
        Matplotlib figure to render in HTML.

    Returns
    -------
    str
        Rendered HTML <svg> tag with styling.
    """

    str_i = StringIO()
    fig.savefig(str_i, format="svg", transparent=True, bbox_inches="tight")
    fig.canvas.draw_idle()
    svg = str_i.getvalue()

    # Include some CSS in the SVG to render nicely in Robot report's dark and
    # light modes
    return "\n".join(
        [
            "<svg",
            'version="1.1"',
            'baseProfile="full"',
            'width="810" height="540" viewBox="0 0 810 540"'
            'xmlns="http://www.w3.org/2000/svg">',
            "<style>",
            "    path {",
            "        fill: var(--text-color);",
            "        stroke: var(--text-color);",
            "    }",
            "</style>",
            svg,
            "</svg>",
        ]
    )


def reference_converter(
    value: PathLike | xr.Dataset | None,
) -> xr.Dataset | None:
    """
    A converter for handling the reference data attribute.

    Parameters
    ----------
    value : path-like or Dataset or None
        Path to the reference dataset file or dataset identifier or Dataset or
        None.

    Raises
    ------
    ValueError
        If the reference data is not a valid dataset.

    Returns
    -------
    xr.Dataset or None
        The reference dataset.

    Notes
    -----
    * ``None`` and datasets are passed through.
    * If ``value`` is a path, resolve it with the path resolver and try to load it.
    * If ``value`` is a path to a file that does not exist, return ``None``.
    """
    if value is None:
        return value

    if isinstance(value, xr.Dataset):
        return value

    if isinstance(value, (str, os.PathLike, bytes)):
        try:
            logger.info(f'Looking up "{str(value)}" on disk', also_console=True)
            fname = fresolver.resolve(value)
            logger.info(f"Resolved path: {fname}", also_console=True)
            return xr.load_dataset(fname)

        except (DataError, FileNotFoundError):
            pass

        # File not found: most likely means reference data does not exist
        return None

    raise ValueError(
        "Reference must be provided as a Dataset, a file path or None. "
        f"Got {type(value).__name__}"
    )


@define
class RegressionTest(ABC):
    """
    Common interface for tests based on the comparison of a result array against
    reference values.
    """

    # Name used for the reference metric. Must be set be subclasses.
    METRIC_NAME: ClassVar[str | None] = None

    name: str = documented(
        attrs.field(validator=attrs.validators.instance_of(str)),
        doc="Test case name.",
        type="str",
        init_type="str",
    )

    value: xr.Dataset = documented(
        attrs.field(
            validator=attrs.validators.instance_of(xr.Dataset),
            repr=summary_repr,
        ),
        doc="Simulation result. Must be specified as a dataset.",
        type=":class:`xarray.Dataset`",
        init_type=":class:`xarray.Dataset`",
    )

    reference: xr.Dataset | None = documented(
        attrs.field(
            default=None,
            converter=reference_converter,
            validator=attrs.validators.optional(
                attrs.validators.instance_of(xr.Dataset)
            ),
        ),
        doc="Reference data. Can be specified as an xarray dataset, a path to a "
        "NetCDF file or a path to a resource.",
        type=":class:`xarray.Dataset` or None",
        init_type=":class:`xarray.Dataset` or path-like, optional",
        default="None",
    )

    variable: str = documented(
        attrs.field(kw_only=True, default="brf_srf"),
        doc="Tested variable",
        type="str",
        init_type="str",
        default="brf_srf",
    )

    threshold: float = documented(
        attrs.field(kw_only=True),
        doc="Test metric threshold",
        type="float",
        init_type="float",
    )

    archive_dir: Path = documented(
        attrs.field(kw_only=True, converter=lambda x: Path(x).resolve()),
        doc="Path to output artefact storage directory. Relative paths are "
        "interpreted with respect to the current working directory.",
        type=":class:`pathlib.Path`",
        init_type="path-like",
    )

    def __attrs_pre_init__(self):
        if self.METRIC_NAME is None:
            raise TypeError(f"Unsupported test type {type(self).__name__}")

    def run(self) -> bool:
        """
        This method controls the execution steps of the regression test:

        * handle missing reference data
        * catch errors during text evaluation
        * create the appropriate plots and data archives

        Returns
        -------
        bool
            Result of the test criterion comparison.
        """

        logger.info(f"Regression test {self.name} results:", also_console=True)

        fname = self.name
        ext = ".nc"
        archive_dir = self.archive_dir

        fname_reference = archive_dir / f"{fname}-ref{ext}"
        fname_result = archive_dir / f"{fname}-result{ext}"

        # if no valid reference is found, store the results as new ref and fail
        # the test
        if not self.reference:
            logger.info(
                "No reference data found. Storing test results to "
                f"{fname_reference}. This can be the new reference.",
                also_console=True,
            )
            self._archive(self.value, fname_reference)
            self._plot(metric_value=None, noref=True)
            return False

        # else (we have a reference value), evaluate the test metric
        try:
            passed, metric_value = self._evaluate()
            msg = "\n".join(
                [
                    "Test passed" if passed else "Test did not pass",
                    f"Metric value: {self.METRIC_NAME} = {metric_value}",
                    f"Metric threshold: {self.threshold}",
                    f"Variable: {self.variable}",
                ]
            )
            logger.info(msg, also_console=True)

        except Exception as e:
            logger.info(
                "An exception occurred during test evaluation!", also_console=True
            )
            self._plot(noref=False, metric_value=None)
            raise e

        # we got a metric: report the results in the archive directory
        logger.info(
            f"Saving current output dataset to {fname_result}", also_console=True
        )
        self._archive(self.value, fname_result)
        logger.info(
            f"Saving reference dataset locally to {fname_reference}", also_console=True
        )
        self._archive(self.reference, fname_reference)
        self._plot(noref=False, metric_value=metric_value)

        return passed

    @abstractmethod
    def _evaluate(self, diagnostic_chart: bool = False) -> tuple[bool, float]:
        """
        Evaluate the test results and perform a comparison to the reference
        based on the criterion defined in the specialized class.

        Parameters
        ----------
        diagnostic_chart : bool, optional
            If ``True``, append a diagnostic chart to the test report.

        Returns
        -------
        passed : bool
            ``True`` iff the test passed.

        metric_value : float
            The value of the test metric.
        """
        pass

    def _archive(self, dataset: xr.Dataset, fname_output: PathLike) -> None:
        """
        Create an archive file for test result and reference storage
        """
        os.makedirs(os.path.dirname(fname_output), exist_ok=True)
        dataset.to_netcdf(fname_output)

    def _plot(self, metric_value: float | None, noref: bool) -> None:
        """
        Create a plot to visualize the results of the test.
        If the ``reference only`` parameter is set, create only a simple plot
        visualizing the new reference data. Otherwise, create the more complex
        comparison plots for the regression test.

        Parameters
        ----------
        metric_value : float or None
            The numerical value of the test metric.

        noref : bool
            If ``True``, create only a simple visualization of the computed
            data.
        """

        fname = self.name
        ext = ".png"

        archive_dir = self.archive_dir
        fname_plot = archive_dir / f"{fname}{ext}"
        fname_plot.parent.mkdir(parents=True, exist_ok=True)

        if noref:
            figure, _ = self._plot_noref()
        else:
            figure, _ = self._plot_ref(metric_value)

        html_svg = figure_to_html(figure)
        logger.info(html_svg, html=True, also_console=False)
        logger.info(f"Saving PNG report chart to {fname_plot}", also_console=True)

        plt.savefig(fname_plot)
        plt.close()

    def _plot_noref(self):
        """
        Draw a simple plot when no reference data is available.
        """
        vza = np.squeeze(self.value.vza.values)
        val = np.squeeze(self.value[self.variable].values)

        fig, ax = plt.subplots(1, 1, figsize=(8, 6))
        ax.plot(vza, val)
        ax.set_xlabel("VZA [deg]")
        ax.set_ylabel(self.variable)
        ax.set_title("Simulation result, can be used as new reference")

        return fig, ax

    def _plot_ref(self, metric_value: float | None = None):
        """
        Draw a comparison plot with reference and test data displayed together.
        """
        vza = np.squeeze(self.value.vza.values)
        val = np.squeeze(self.value[self.variable].values)
        ref = np.squeeze(self.reference[self.variable].values)

        return regression_test_plots(
            ref,
            val,
            vza,
            (self.METRIC_NAME, metric_value),
            xlabel="VZA [deg]",
            ylabel=self.variable,
        )

    def _plot_diagnostic(self, **diagnostic_info) -> None:
        """
        Create an additional plot to display more technical information about
        the test metrics and decision process. The diagnostic plot can help the
        user debug a failing test, or to assess the test power and significance.
        This plot is output directly to the robotframework report.

        Parameters:
        -----------
        **diagnostic_info: dict
            Variadic keyword arguments for the subclasses implementation
        """

        raise NotImplementedError(
            f"{type(self)} does not implement a diagnostic plot method"
        )


@define
class RMSETest(RegressionTest):
    """
    This class implements a simple test based on the root mean squared
    error (RMSE) of a result array with respect to the reference data.

    The test will pass if the computed root mean squared deviation between
    the result and reference is smaller or equal to the given threshold.
    """

    METRIC_NAME = "rmse"

    def _evaluate(self) -> tuple[bool, float]:
        value_np = self.value.brf.values
        ref_np = self.reference.brf.values
        if np.shape(value_np) != np.shape(ref_np):
            raise ValueError(
                f"Result and reference do not have the same shape! "
                f"Got: {np.shape(value_np)}, {np.shape(ref_np)}"
            )

        result_flat = np.array(value_np).flatten()
        ref_flat = np.array(ref_np).flatten()

        rmse = np.linalg.norm(result_flat - ref_flat) / np.sqrt(len(ref_flat))
        return rmse <= self.threshold, rmse


@define
class Chi2Test(RegressionTest):
    """
    This class implements a statistical test for the regression testing
    campaign, based on Pearson's Chi-squared test.
    https://en.wikipedia.org/wiki/Pearson%27s_chi-squared_test

    It determines the probability for the reference and the test result
    following the same distribution.

    This test will pass if the computed p-value is strictly larger than the
    given threshold.
    """

    # The algorithm is adapted from Mitsuba's testing framework.

    METRIC_NAME = "X² p-value"

    def _evaluate(self, diagnostic_chart=False) -> tuple[bool, float]:
        ref_np = self.reference.brf.values

        result_np = self.value.brf.values
        histo_bins = np.linspace(ref_np.min(), ref_np.max(), 20)
        histo_ref = np.histogram(ref_np, histo_bins)[0]
        histo_res = np.histogram(result_np, histo_bins)[0]

        # sorting both histograms following the ascending frequencies in
        # the reference. Algorithm from:
        # https://stackoverflow.com/questions/9764298/how-to-sort-two-lists-which-reference-each-other-in-the-exact-same-way
        histo_ref_sorted, histo_res_sorted = zip(
            *sorted(zip(histo_ref, histo_res), key=lambda x: x[0])
        )

        from mitsuba.python.math import rlgamma

        chi2val, dof, pooled_in, pooled_out = mi.math.chi2(
            histo_res_sorted, histo_ref_sorted, 5
        )
        p_value = 1.0 - rlgamma(dof / 2.0, chi2val / 2.0)

        return p_value > self.threshold, p_value


@define
class AbstractStudentTTest(RegressionTest):
    """
    Abstract Student's T-Test
    =========================

    Implement diagnostic chart common to subclassing T-test implementations.
    """

    def _plot_diagnostic(self, dof=None, t_prim=None) -> None:
        """
        Diagnostic chart for an Independant Student's T-test

        Parameters:
        -----------
        dof: int
            Degrees of Freedom
        t_prim: float
            t' statistic issued from the test
        """
        fig, ax = plt.subplots()
        ax.grid()
        ax2 = ax.twinx()

        start, end = spstats.t.ppf(0.0001, dof), spstats.t.ppf(0.9999, dof)

        if (t_prim > start) and (t_prim < end):
            fx = np.linspace(-np.abs(t_prim), np.abs(t_prim), 100)
            fy = spstats.t(dof).pdf(fx)
            ax.fill_between(np.zeros((100,)), fy)
            ax.axvline(spstats.t.ppf(-self.threshold / 2.0, dof), color="red")
            ax.axvline(spstats.t.ppf(self.threshold / 2.0, dof), color="red")
        else:
            ax.axvline(t_prim, label="T value")

        ax.axvline(0.0, color="red", linestyle="--")
        ax.set_title("T-statistic")

        x = np.linspace(start, end, 100)
        y = spstats.t(dof).pdf(x)

        ax2.plot(x, y, label="target T distribution form", color="black")
        ax2.legend(loc="upper right")
        ax2.set_ylim([0.0, max(y) * 1.1])

        chart = figure_to_html(fig)
        plt.close(fig)

        logger.info(chart, html=True)


@define
class IndependantStudentTTest(AbstractStudentTTest):
    """
    Independant Student's T-test
    ============================

    This implementation of a Student's T-test is following the assumption of
    independance of the two groups that are tested. The bias of the mean values
    of the two groups is assumed to be the result of chance under the null
    hypothesis. It is a two-tailed test.

    It is less sensitive to outliers than the paired Student's T-test.
    """

    METRIC_NAME = "T-test p-value"

    def _evaluate(self, diagnostic_chart=False) -> tuple[bool, float]:
        variable_var = self.variable + "_var"

        if variable_var not in self.reference:
            raise ValueError(
                "The reference data for this T-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        if variable_var not in self.value:
            raise ValueError(
                "The tested data for this T-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        ref_np = self.reference[self.variable].values.ravel()
        result_np = self.value[self.variable].values.ravel()

        var_ref_np = self.reference[variable_var].values.ravel()
        var_res_np = self.value[variable_var].values.ravel()

        # Calculate mean values over observations and associated variances
        R_res = np.mean(result_np)
        R_ref = np.mean(ref_np)
        var_R_res = np.sum(var_res_np) / var_res_np.size**2
        var_R_ref = np.sum(var_ref_np) / var_ref_np.size**2
        bias_mean = R_res - R_ref

        # Calculate T-statistic and associated degree of freedom of its
        # T-distribution using a pooled standard deviation

        s_p = np.sqrt(
            ((var_res_np.size - 1.0) * var_R_res + (var_ref_np.size - 1.0) * var_R_ref)
            / (var_res_np.size + var_ref_np.size - 2)
        )
        t_prim = bias_mean / (
            s_p * np.sqrt(1.0 / var_res_np.size + 1.0 / var_ref_np.size)
        )

        dof = (var_res_np.size + var_ref_np.size) - 2

        assert dof > 0

        # Calculate p-value of the two-tailed t-test using the T distribution
        # survival function for the null hypothesis.
        p_value = spstats.t.sf(np.abs(t_prim), dof) * 2

        passed = p_value > self.threshold

        if diagnostic_chart:
            self._plot_diagnostic(dof=dof, t_prim=t_prim)

        logger.info(f"bias    = {bias_mean}", also_console=True)
        logger.info(f"s_p     = {s_p}", also_console=True)
        logger.info(f"t'      = {t_prim}", also_console=True)
        logger.info(f"dof     = {dof}", also_console=True)
        logger.info(f"p-value = {p_value}", also_console=True)
        logger.info(f"alpha   = {self.threshold}", also_console=True)

        return passed, p_value


@define
class PairedStudentTTest(AbstractStudentTTest):
    """
    Paired Student's T-test
    =======================

    This implementation of a Student's T-test is following the assumption of
    paired samples within two groups that are tested. The mean of the bias
    between the paired values is assumed to be the result of chance under the
    null hypothesis. It is a two-tailed test.

    The paired test allow to introduce a covariance factor between the pairs.
    By default, this covariance is equal to zero, thus assuming independance of
    the two variables.

    Contrary to the independant Student's T-test, this paired version of the
    test requires an equal degree of freedom of the two groups.
    """

    METRIC_NAME = "paired T-test p-value"

    cov: np.ndarray | float = documented(
        attrs.field(kw_only=True, default=0.0),
        doc="Covariance between observation, defaults to zero",
        type="ndarray or float",
        init_type="array-like or float",
    )

    def _evaluate(self, diagnostic_chart=False) -> tuple[bool, float]:
        variable_var = self.variable + "_var"

        if variable_var not in self.reference:
            raise ValueError(
                "The reference data for this T-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        if variable_var not in self.value:
            raise ValueError(
                "The tested data for this T-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        ref_np = self.reference[self.variable].values.ravel()
        result_np = self.value[self.variable].values.ravel()

        var_ref_np = self.reference[variable_var].values.ravel()
        var_res_np = self.value[variable_var].values.ravel()

        assert ref_np.shape == result_np.shape
        assert ref_np.shape == var_ref_np.shape
        assert ref_np.shape == var_res_np.shape

        # Calculate paired mean value and associated variance
        D_mean = np.mean(result_np - ref_np)
        var_D = (var_res_np + var_ref_np) - 2 * self.cov
        var_D_mean = np.sum(var_D) / var_D.size**2

        # Calculate T-statistic and associated degree of freedom of its
        # T-distribution
        t_prim = D_mean / (var_D_mean / np.sqrt(var_D.size))
        dof = var_D.size - 1

        assert dof > 0

        # Calculate p-value of the two-tailed t-test using the T distribution
        # survival function for the null hypothesis.
        p_value = spstats.t.sf(np.abs(t_prim), dof) * 2

        passed = p_value > self.threshold

        if diagnostic_chart:
            self._plot_diagnostic(dof=dof, t_prim=t_prim)

        logger.info(f"bias     = {D_mean}", also_console=True)
        logger.info(f"var mean = {var_D_mean}", also_console=True)
        logger.info(f"t'       = {t_prim}", also_console=True)
        logger.info(f"dof      = {dof}", also_console=True)
        logger.info(f"p-value  = {p_value}", also_console=True)
        logger.info(f"alpha    = {self.threshold}", also_console=True)

        return passed, p_value


@define
class ZTest(RegressionTest):
    """
    Z-Test with Šidák correction factor
    ===================================

    Implement a Z-test, testing the significance of paired differences between
    a set of observations and a set of references. It considers the observations
    variance.

    Paired tests are aggregated into one p-value using a Šidák correction. The
    test passes if the null hypothesis is accepted for at least 99.75% of the
    paired Z-tests

    This paired Z-test requires an equal degree of freedom of the two groups.
    """

    METRIC_NAME = "Z-test p-value"

    def _plot_diagnostic(self, z=None) -> None:
        """
        Diagnostic chart for a Z-test

        Parameters:
        -----------
        z: array-like
            Z-statistic for each pair of measurements
        """

        fig, ax = plt.subplots()
        ax.grid()
        ax2 = ax.twinx()

        ax.hist(z, bins=50, label="Z values")
        ax.axvline(0.0, color="red", linestyle="--")
        ax.set_title("Z-statistic")
        ax.legend(loc="upper left")

        x = np.linspace(-4.0, 4.0, 100)
        y = spstats.norm.pdf(x, 0.0, 1.0)
        ax2.plot(x, y, label="target Z distribution form", color="black")
        ax2.legend(loc="upper right")
        ax2.set_ylim([0.0, max(y) * 1.1])

        chart = figure_to_html(fig)
        plt.close(fig)

        logger.info(chart, html=True)

    def _evaluate(self, diagnostic_chart=False) -> tuple[bool, float]:
        variable_var = self.variable + "_var"

        if variable_var not in self.value:
            raise ValueError(
                "The reference data for this Z-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        ref_np = self.reference[self.variable].values.ravel()
        result_np = self.value[self.variable].values.ravel()

        var_res_np = self.value[variable_var].values.ravel()

        assert ref_np.shape == result_np.shape
        assert ref_np.shape == var_res_np.shape

        # Calculate Z-statistic
        z = (result_np - ref_np) / np.sqrt(var_res_np)

        # Calculate p-value of the two-tailed z-test null hypothesis
        p_values = spstats.norm.sf(np.abs(z)) * 2

        alpha_0 = 1.0 - (1.0 - self.threshold) ** (1.0 / result_np.size)
        accept_null = p_values > alpha_0

        passed = np.count_nonzero(accept_null) >= 0.9975 * result_np.size

        if diagnostic_chart:
            self._plot_diagnostic(z=z)

        logger.info(f"min p-value = {min(p_values)}", also_console=True)
        logger.info(f"max p-value = {max(p_values)}", also_console=True)
        logger.info(
            f"n passed    = {np.count_nonzero(accept_null)}/{0.99 * result_np.size}",
            also_console=True,
        )
        logger.info(f"alpha_1     = {self.threshold}", also_console=True)
        logger.info(f"alpha_0     = {alpha_0}", also_console=True)

        return passed, min(p_values)

    def _plot_ref(self, metric_value: float | None = None):
        """
        Draw a comparison plot with reference and test data displayed together.
        """
        vza = np.squeeze(self.value.vza.values)
        result = np.squeeze(self.value[self.variable].values)
        result_var = np.squeeze(self.value[f"{self.variable}_var"].values)
        ref = np.squeeze(self.reference[self.variable].values)

        return regression_test_plots(
            ref,
            result,
            vza,
            (self.METRIC_NAME, metric_value),
            result_var=result_var,
            xlabel="VZA [deg]",
            ylabel=self.variable,
        )


@define
class SidakTTest(RegressionTest):
    """
    T-Test with Šidák correction factor
    ===================================

    Implement a T-test, testing the significance of paired differences between
    a set of observations and a set of references. It considers both the
    observations and reference variance.

    Paired tests are aggregated into one p-value using a Šidák correction. The
    test passes if the null hypothesis is accepted for at least 99.75% of the
    paired T-tests
    """

    METRIC_NAME = "Sidak T-test p-value"

    def _plot_diagnostic(self, t_prim=None) -> None:
        """
        Diagnostic chart for the T-test

        Parameters:
        -----------
        t_prim: array-like
            T-statistic for each pair of measurements
        """

        fig, ax = plt.subplots()
        ax.grid()
        ax2 = ax.twinx()

        start, end = spstats.norm.ppf(0.0001), spstats.norm.ppf(0.9999)

        ax.hist(t_prim, bins=50, label="T values")
        ax.axvline(0.0, color="red", linestyle="--")
        ax.set_title("T-statistic")
        ax.legend(loc="upper left")

        x = np.linspace(start, end, 100)
        y = spstats.norm.pdf(x)
        ax2.plot(x, y, label="target T distribution form", color="black")
        ax2.legend(loc="upper right")
        ax2.set_ylim([0.0, max(y) * 1.1])

        chart = figure_to_html(fig)
        plt.close(fig)

        logger.info(chart, html=True)

    def _evaluate(self, diagnostic_chart=False) -> tuple[bool, float]:
        variable_var = self.variable + "_var"

        if variable_var not in self.reference:
            raise ValueError(
                "The reference data for this T-test does not contain expected "
                "appropriate variance values, could not find data variable "
                f"'{variable_var}'"
            )

        ref_np = self.reference[self.variable].values.ravel()
        result_np = self.value[self.variable].values.ravel()

        assert ref_np.shape == result_np.shape

        var_ref_np = self.reference[variable_var].values.ravel()
        var_res_np = self.value[variable_var].values.ravel()

        assert var_ref_np.shape == var_res_np.shape

        # Calculate T-statistic
        t_prim = (result_np - ref_np) / np.sqrt(var_res_np + var_ref_np)

        # Calculate p-value of the two-tailed t-test using the T distribution
        # survival function for the null hypothesis that there is no difference
        # between the two mean distributions. It is assumed that the sample size
        # is large enough for the T distribution to converge to a normal one.
        p_values = spstats.norm.sf(np.abs(t_prim)) * 2

        # Calculate the Šidák correction
        alpha_0 = 1.0 - (1.0 - self.threshold) ** (1.0 / result_np.size)
        accept_null = p_values > alpha_0

        passed = np.count_nonzero(accept_null) >= 0.9975 * result_np.size

        if diagnostic_chart:
            self._plot_diagnostic(t_prim=t_prim)

        logger.info(f"min p-value = {min(p_values)}", also_console=True)
        logger.info(f"max p-value = {max(p_values)}", also_console=True)
        logger.info(
            f"n passed    = {np.count_nonzero(accept_null)}/{0.99 * result_np.size}",
            also_console=True,
        )
        logger.info(f"alpha_1     = {self.threshold}", also_console=True)
        logger.info(f"alpha_0     = {alpha_0}", also_console=True)

        return passed, min(p_values)



from __future__ import annotations

import os
import typing as t
import warnings
from abc import ABC, abstractmethod

import numpy as np
import numpy.typing as npt
import pint
import pinttrs
import scipy.integrate as spi
import xarray as xr
from pinttrs.util import ensure_units

from .. import converters, validators
from ..attrs import define, documented
from ..data import fresolver
from ..exceptions import DataError
from ..units import to_quantity
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg
from ..util.misc import summary_repr

# ------------------------------------------------------------------------------
#                             Class implementations
# ------------------------------------------------------------------------------


@define
class SpectralResponseFunction(ABC):
    """
    An interface defining the spectral response function of an instrument.
    """

    @staticmethod
    def convert(value) -> t.Any:
        """
        Converter for the ``Measure.srf`` field.

        Notes
        -----
        Supported conversion protocols:

        * :class:`dict`: Dispatch to subclass based on 'type' entry, then pass
          dictionary to constructor as keyword arguments.
        * Dataset, DataArray: Call :meth:`.BandSRF.from_dataarray`.
        * Path-like: Attempt loading a dataset from the hard drive, then call
          :meth:`.BandSRF.from_dataarray`.
        * :class:`str`: Perform a NetCDF file lookup in the SRF database and load
          it.

        Anything else will pass through this converter without modification.
        """
        if isinstance(value, dict):
            d = value.copy()
            try:
                type_id = d.pop("type")
            except KeyError as e:
                raise ValueError(
                    "missing 'type' key in SRF specification dictionary"
                ) from e

            dispatch_table = {
                "uniform": UniformSRF,
                "delta": DeltaSRF,
                "multi_delta": DeltaSRF,
                "band": BandSRF,
            }

            try:
                cls = dispatch_table[type_id]
            except KeyError as e:
                raise ValueError(f"unknown SRF type '{type_id}'") from e

            return cls(**d)

        if isinstance(value, xr.Dataset):
            return BandSRF.from_dataarray(value.srf)

        if isinstance(value, xr.DataArray):
            return BandSRF.from_dataarray(value)

        if isinstance(value, (str, os.PathLike)):
            try:
                ds = xr.load_dataset(value)
                return BandSRF.from_dataarray(ds.srf)
            except (FileNotFoundError, ValueError):
                pass

        if isinstance(value, str):
            try:
                return BandSRF.from_id(value)
            except DataError:
                pass

        return value

    @abstractmethod
    def eval(self, w: npt.ArrayLike) -> pint.Quantity:
        """
        Evaluate the spectral response function for one or several wavelengths.
        Evaluation is vectorized.

        Parameters
        ----------
        w : array-like
            One or several wavelengths at which the SRF is evaluated.

        Returns
        -------
        quantity
            The returned value as the same shape as ``w``.
        """
        pass


@define
class UniformSRF(SpectralResponseFunction):
    r"""
    A spectral response function uniform on a preset spectral interval.

    It represents a function defined by:

    .. math::

       f(w) = \left\{
       \begin{array}{r l}
            \text{value} & \text{if} \ w \in [w_\mathrm{min}, w_\mathrm{max}] \\
            0 & \text{otherwise}
        \end{array}
        \right.
    """

    wmin = documented(
        pinttrs.field(
            units=ucc.deferred("wavelength"),
            factory=lambda: 300.0 * ureg.nm,
            validator=validators.is_positive,
        ),
        doc="Lower bound of the interval.",
        type="quantity",
        init_type="quantity or float",
        default="300 nm",
    )

    wmax = documented(
        pinttrs.field(
            units=ucc.deferred("wavelength"),
            factory=lambda: 2500.0 * ureg.nm,
            validator=validators.is_positive,
        ),
        doc="Upper bound of the interval.",
        type="quantity",
        init_type="quantity or float",
        default="2500 nm",
    )

    value = documented(
        pinttrs.field(
            units=ucc.deferred("dimensionless"),
            factory=lambda: 1.0,
            validator=validators.is_positive,
        ),
        doc="The value to which the phase function evaluates in its interval.",
        type="quantity",
        init_type="quantity or float",
        default="1.0",
    )

    def plot(self, ax, alpha=0.5):
        w_u = ucc.get("wavelength")
        x = [self.wmin.m_as(w_u), self.wmax.m_as(w_u)]
        y = [self.value.m, self.value.m]

        ax.fill_between(x, y, alpha=alpha)
        ax.plot(x, y, marker=".")
        ax.vlines(x, 0, 1)

        return ax

    def _repr_html_(self):
        import base64
        import io

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1, figsize=(6, 1.5))

        self.plot(ax)
        w_u = ucc.get("wavelength")
        ax.set_xlabel(f"Wavelength [{w_u:~P}]")

        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        plt.close(fig)
        img.seek(0)

        return (
            "<img "
            f'src="data:image/png;base64, {base64.b64encode(img.getvalue()).decode("utf-8")}" '
            "/>"
        )

    def eval(self, w: npt.ArrayLike) -> pint.Quantity:
        # Inherit docstring

        w_units = ucc.get("wavelength")
        w_m = pinttrs.util.ensure_units(w, w_units).m_as(w_units)

        return (
            np.where(
                (w_m >= self.wmin.m_as(w_units)) & (w_m <= self.wmax.m_as(w_units)),
                self.value.m,
                np.zeros_like(w),
            )
            * self.value.u
        )


@define
class DeltaSRF(SpectralResponseFunction):
    """
    A spectral response function consisting of multiple Dirac delta distributions.

    This SRF conventionally always evaluates to 0.
    """

    wavelengths: pint.Quantity = documented(
        pinttrs.field(
            units=ucc.deferred("wavelength"),
            converter=[
                converters.on_quantity(np.atleast_1d),
                converters.on_quantity(lambda x: x.astype(np.float64)),
                pinttrs.converters.to_units(ucc.deferred("wavelength")),
            ],
            validator=validators.all_strictly_positive,
            repr=summary_repr,
        ),
        doc="An array of wavelengths specifying the translation wavelength of each "
        "Dirac delta. Wavelength values are positive and unique. "
        "When a single value is provided, it is converted to a 1-element array. "
        "Wavelength are deduplicated and sorted by ascending values. "
        'Unit-enabled field (default: ``ucc["wavelength"]``).',
        type="quantity",
        init_type="array-like or quantity",
    )

    def plot(self, ax):
        ax.vlines(self.wavelengths.m, 0, 1)
        return ax

    def _repr_html_(self):
        import base64
        import io

        import matplotlib.pyplot as plt
        import seaborn as sns

        fig, ax = plt.subplots(1, 1, figsize=(6, 1))
        self.plot(ax)
        ax.set_xlabel(f"Wavelength [{self.wavelengths.u:~P}]")
        sns.despine(left=True)
        ax.axes.get_yaxis().set_visible(False)

        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        plt.close(fig)
        img.seek(0)

        return (
            "<img "
            f'src="data:image/png;base64, {base64.b64encode(img.getvalue()).decode("utf-8")}" '
            "/>"
        )

    def eval(self, w: npt.ArrayLike) -> pint.Quantity:
        # Inherit docstring

        return np.zeros_like(w) * ureg.dimensionless


@define
class BandSRF(SpectralResponseFunction):
    """
    The spectral response function for a single band of an instrument.
    """

    wavelengths: pint.Quantity = documented(
        pinttrs.field(
            units=ucc.deferred("wavelength"),
            converter=[
                converters.on_quantity(np.atleast_1d),
                pinttrs.converters.to_units(ucc.deferred("wavelength")),
            ],
            validator=validators.is_sorted(strict=True),
            repr=summary_repr,
        ),
        doc="Wavelengths defining the interpolation grid. Values must be "
        "monotonically increasing.",
        type="quantity",
    )

    values: pint.Quantity = documented(
        pinttrs.field(
            units=ucc.deferred("dimensionless"),
            converter=[
                converters.on_quantity(np.atleast_1d),
                pinttrs.converters.to_units(ucc.deferred("dimensionless")),
            ],
        ),
        doc="Spectrum values. If a quantity is passed, units are "
        "checked for consistency. If a unitless array is passed, it is "
        "automatically converted to appropriate default configuration units, "
        "depending on the value of the ``quantity`` field. If no quantity is "
        "specified, this field can be a unitless value.",
        type="quantity",
        init_type="array-like",
    )

    @values.validator
    @wavelengths.validator
    def _values_wavelengths_validator(self, attribute, value):
        # Check size
        if value.ndim > 1:
            raise ValueError(
                f"while validating '{attribute.name}': '{attribute.name}' must "
                "be a 1D array"
            )

        if len(value) < 2:
            raise ValueError(
                f"while validating '{attribute.name}': '{attribute.name}' must "
                f"have length >= 2"
            )

        if self.wavelengths.shape != self.values.shape:
            raise ValueError(
                f"while validating '{attribute.name}': 'wavelengths' and 'values' "
                f"must have the same shape, got {self.wavelengths.shape} and "
                f"{self.values.shape}"
            )

    @values.validator
    def _values_validator(self, attribute, value):
        # Check for leading and trailing zeros
        if not np.all(value.m[[0, -1]] == np.array([0.0, 0.0])):
            warnings.warn(
                "Initializing a BandSRF instance without a leading and trailing "
                "zero is not recommended."
            )

    @classmethod
    def from_dataarray(cls, da: xr.DataArray):
        return cls(wavelengths=to_quantity(da.w), values=da.values)

    @classmethod
    def from_id(cls, id: str):
        fname = fresolver.resolve(f"srf/{id}.nc")
        try:
            ds = xr.load_dataset(fname)
        except FileNotFoundError as e:
            raise DataError(f"could not load SRF with identifier '{id}'") from e

        return cls.from_dataarray(ds["srf"])

    def plot(self, ax, alpha=0.5, lw=1):
        w_u = ucc.get("wavelength")
        x = self.wavelengths.m_as(w_u)
        y = self.values.m
        if alpha:
            ax.fill_between(x, y, alpha=alpha)
        ax.plot(x, y, lw=lw, marker=".")
        return ax

    def _repr_html_(self):
        import base64
        import io

        import matplotlib.pyplot as plt

        w_u = ucc.get("wavelength")
        fig, ax = plt.subplots(1, 1, figsize=(6, 1.5))

        self.plot(ax)
        ax.set_xlabel(f"Wavelength [{w_u:~P}]")

        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        plt.close(fig)
        img.seek(0)

        return (
            "<img "
            f'src="data:image/png;base64, {base64.b64encode(img.getvalue()).decode("utf-8")}" '
            "/>"
        )

    def support(self) -> pint.Quantity:
        """
        Return the interval in which the SRF is nonzero.

        Returns
        -------
        quantity

        Note
        ----
        This method does not actually compute the support: it assumes that any
        leading and trailing zeros are already removed.
        """
        wmin_m = self.wavelengths.m.min()
        wmax_m = self.wavelengths.m.max()
        return (wmin_m, wmax_m) * self.wavelengths.u

    def eval(self, w: npt.ArrayLike) -> pint.Quantity:
        # Inherit docstring

        w_units = self.wavelengths.u
        w_m = pinttrs.util.ensure_units(w, ucc.get("wavelength")).m_as(w_units)
        return (
            np.interp(w_m, self.wavelengths.m, self.values.m, left=0.0, right=0.0)
            * self.values.u
        )

    def integrate(
        self, wmin: float | pint.Quantity, wmax: float | pint.Quantity
    ) -> pint.Quantity:
        """
        Return the integral of the SRF on the specified interval, using the
        trapezoid rule.

        Parameters
        ----------
        wmin, wmax : float or quantity
            Lower and upper bounds of the integration domain. Floats are
            interpreted as being specified in default configuration units.

        Returns
        -------
        integral : quantity
            Integral as a scalar quantity.
        """
        wmin = ensure_units(wmin, ucc.get("wavelength"))
        wmax = ensure_units(wmax, ucc.get("wavelength"))

        # Assemble spectral integration mesh
        w_u = self.wavelengths.u
        w_m = np.unique(
            np.concatenate(
                (
                    np.atleast_1d(wmin.m_as(w_u)),
                    self.wavelengths.m[
                        (self.wavelengths >= wmin) & (self.wavelengths <= wmax)
                    ],
                    np.atleast_1d(wmax.m_as(w_u)),
                )
            )
        )

        # Evaluate SRF at mesh nodes
        values_m = self.eval(w_m * w_u).m

        # Compute integral
        return spi.trapezoid(values_m, w_m) * w_u

    def integrate_cumulative(self, w: npt.ArrayLike) -> pint.Quantity:
        """
        Return the cumulative integral of the SRF on the specified mesh, using
        the trapezoid rule.

        Parameters
        ----------
        w : array-like
            Nodes of the spectral integration mesh. If a dimensionless array is
            passed, it is interpreted as being specified in default
            configuration units.

        Returns
        -------
        integral : quantity
            Cumulative integral as an array of shape (N-1,), where
            ``wavelength`` has shape (N,).
        """
        w_u = ucc.get("wavelength")
        w_m = ensure_units(w, w_u).m_as(w_u)

        # Evaluate SRF at mesh nodes
        values_m = self.eval(w).m

        # Compute integral
        return spi.cumulative_trapezoid(values_m, w_m) * w_u


@ureg.wraps(ret=None, args=("nm", "nm", None, "nm", "nm", None, None), strict=False)
def make_gaussian(
    wl_center: pint.Quantity,
    fwhm: pint.Quantity,
    cutoff: float = 3.0,
    wl: pint.Quantity | None = None,
    wl_res: pint.Quantity | float = 1.0,
    pad: bool = False,
    normalize: bool = True,
) -> xr.Dataset:
    """
    Generate a Gaussian spectral response function dataset from central
    wavelength and full width at half maximum values.

    Parameters
    ----------
    wl_center : quantity or float
        Central wavelength of the Gaussian distribution.
        If passed as a float, the value is interpreted as being given in nm.

    fwhm : quantity or float
        Full width at half maximum of the Gaussian distribution.
        If passed as a float, the value is interpreted as being given in nm.

    cutoff : float, default: 3.0
        Cut-off, in multiples of the standard deviation σ.

    wl : quantity or array-like, optional
        Mesh used to evaluate the discretized distribution. If unset, a regular
        mesh with spacing given by ``wl_res`` is used.
        If passed as an array, the value is interpreted as being given in nm.

    wl_res : quantity or float, optional
        Resolution of the automatic spectral mesh if relevant.
        If passed as a float, the value is interpreted as being given in nm.

    pad : bool, default: False
        If True, pad SRF data with leading and trailing zeros.

    normalize : bool, default: True
        If ``True``, the generated SRF data is normalized to have a maximum
        equal to 1.

    Returns
    -------
    Dataset
        A dataset compliant with the Eradiate SRF format. The uncertainty
        variable is set to NaN.
    """
    # Generate default mesh if necessary
    if wl is None:
        wl = np.arange(0.0, 5001.0, wl_res)

    # Infer standard deviation
    sigma = 0.5 * fwhm / np.sqrt(2.0 * np.log(2.0))

    # Build baseline distribution
    values = np.exp(-0.5 * np.power((wl - wl_center) / sigma, 2)) / (sigma * np.sqrt(2))

    # Prepare cutoff selection
    wl_min = wl_center - cutoff * sigma
    wl_max = wl_center + cutoff * sigma
    wl_mask = (wl >= wl_min) & (wl <= wl_max)

    if pad:  # Apply zero-padding if requested
        i_min = np.argwhere(wl_mask).min() - 1
        i_max = np.argwhere(wl_mask).max() + 1
        wl_mask[[i_min, i_max]] = True
        values_result = values[wl_mask]
        values_result[[0, -1]] = 0.0
    else:  # Otherwise just select
        values_result = values[wl_mask]

    # Build output dataset
    if normalize:
        values_result /= values_result.max()
    wl_result = wl[wl_mask]

    result = xr.Dataset(
        data_vars={
            "srf": (
                ["w"],
                values_result,
                {
                    "standard_name": "spectral_response_function",
                    "long_name": "spectral response function",
                    "units": "dimensionless",
                },
            ),
        },
        coords={
            "w": (
                ["w"],
                wl_result,
                {
                    "standard_name": "radiation_wavelength",
                    "long_name": "wavelength",
                    "units": "nm",
                },
            )
        },
    )
    result["srf_u"] = xr.full_like(result.srf, np.nan)
    result.srf_u.attrs = {
        "standard_name": "spectral_response_function_uncertainty",
        "long_name": "spectral response function uncertainty",
        "units": "dimensionless",
    }

    return result



from __future__ import annotations

from pathlib import Path

import attrs

from ._core import ShapeNode
from ..core import BoundingBox
from ...attrs import define, documented


@define(eq=False, slots=False)
class FileMeshShape(ShapeNode):
    """
    File based mesh shape [``file_mesh``].

    This shape represents a triangulated mesh defined in a file. The OBJ and PLY
    formats are supported.

    Warnings
    --------
    Vertex coordinates are assumed to be defined in kernel units.
    """

    filename: Path = documented(
        attrs.field(converter=Path, kw_only=True),
        type="Path",
        init_type="path-like",
        doc="Path to the mesh file.",
    )

    @filename.validator
    def _filename_validator(self, attribute, value):
        if value.suffix not in {".obj", ".ply"}:
            raise ValueError(
                f"while validating {attribute.name}:"
                f"Eradiate supports mesh files only in PLY or OBJ format."
            )

    def bbox(self) -> BoundingBox:
        # Inherit docstring
        raise NotImplementedError

    @property
    def template(self) -> dict:
        # Inherit docstring

        if self.filename.suffix == ".obj":
            mi_plugin = "obj"
        elif self.filename.suffix == ".ply":
            mi_plugin = "ply"
        else:
            raise ValueError(
                f"unsupported mesh file extension '{self.filename.suffix}'"
            )
        result = {
            "type": mi_plugin,
            "filename": str(self.filename),
            "face_normals": True,
        }
        if self.to_world is not None:
            result["to_world"] = self.to_world

        return result



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod

import attrs

import eradiate

from ..core import NodeSceneElement
from ... import validators
from ..._factory import Factory
from ...attrs import define, documented, get_doc
from ...kernel import DictParameter
from ...spectral.response import DeltaSRF, SpectralResponseFunction
from ...units import unit_registry as ureg

measure_factory = Factory()
measure_factory.register_lazy_batch(
    [
        (
            "_distant.DistantMeasure",
            "distant",
            {},
        ),
        (
            "_distant.MultiPixelDistantMeasure",
            "mpdistant",
            {},
        ),
        (
            "_distant_flux.DistantFluxMeasure",
            "distant_flux",
            {"aliases": ["distantflux"]},
        ),
        (
            "_hemispherical_distant.HemisphericalDistantMeasure",
            "hemispherical_distant",
            {"aliases": ["hdistant"]},
        ),
        (
            "_multi_distant.MultiDistantMeasure",
            "multi_distant",
            {"aliases": ["mdistant"]},
        ),
        (
            "_multi_radiancemeter.MultiRadiancemeterMeasure",
            "multi_radiancemeter",
            {"aliases": ["mradiancemeter"]},
        ),
        (
            "_perspective.PerspectiveCameraMeasure",
            "perspective",
            {},
        ),
        (
            "_radiancemeter.RadiancemeterMeasure",
            "radiancemeter",
            {},
        ),
    ],
    cls_prefix="eradiate.scenes.measure",
)


def _str_summary_raw(x):
    if not x:
        return "{}"

    keys = list(x.keys())

    if len(keys) == 1:
        return f"dict<1 item>({{{keys[0]}: {{...}} }})"
    else:
        return f"dict<{len(keys)} items>({{{keys[0]}: {{...}} , ... }})"


@define(eq=False, slots=False)
class Measure(NodeSceneElement, ABC):
    """
    Abstract base class for all measure scene elements.

    See Also
    --------
    :func:`.mitsuba_run`

    Notes
    -----
    * This class is meant to be used as a mixin.
    * Raw results stored in the `results` field as nested dictionaries with the
      following structure:

      .. code:: python

         {
             spectral_key_0: dict_0,
             spectral_key_1: dict_1,
             ...
         }

      Keys are spectral loop indexes; values are nested dictionaries produced by
      :func:`.mitsuba_run`.
    """

    # --------------------------------------------------------------------------
    #                           Fields and properties
    # --------------------------------------------------------------------------

    id: str | None = documented(
        attrs.field(
            default="measure",
            validator=attrs.validators.optional(attrs.validators.instance_of(str)),
        ),
        doc=get_doc(NodeSceneElement, "id", "doc"),
        type=get_doc(NodeSceneElement, "id", "type"),
        init_type=get_doc(NodeSceneElement, "id", "init_type"),
        default='"measure"',
    )

    mi_results: dict = documented(
        attrs.field(factory=dict, repr=_str_summary_raw, init=False),
        doc="Storage for raw results yielded by the kernel.",
        type="dict",
        default="{}",
    )

    srf: SpectralResponseFunction = documented(
        attrs.field(
            factory=lambda: DeltaSRF(wavelengths=550.0 * ureg.nm),
            converter=SpectralResponseFunction.convert,
            validator=attrs.validators.instance_of(SpectralResponseFunction),
        ),
        doc="Spectral response function (SRF). If a path is passed, it attempts "
        "to load a dataset from that location. If a keyword is passed, *e.g.* "
        "``'sentinel_2a-msi-4'``, the corresponding dataset is looked up "
        "through the file resolver.",
        type=".SpectralResponseFunction",
        init_type="path-like or str or .SpectralResponseFunction or dict",
        default=":class:`DeltaSRF(wavelengths=550.0 * ureg.nm) <.DeltaSRF>`",
    )

    sampler: str = documented(
        attrs.field(
            default="independent",
            validator=attrs.validators.in_(
                {"independent", "stratified", "multijitter", "orthogonal", "ldsampler"}
            ),
        ),
        doc="Mitsuba sampler used to generate pseudo-random number sequences.",
        type="str",
        init_type='{"independent", "stratified", "multijitter", "orthogonal", '
        '"ldsampler"}',
        default='"independent"',
    )

    rfilter: str = documented(
        attrs.field(
            default="box",
            validator=attrs.validators.in_({"box", "gaussian"}),
        ),
        doc="Reconstruction filter used to scatter samples on sensor pixels. "
        "By default, using a box filter is recommended.",
        type="str",
        init_type='{"box", "gaussian"}',
        default='"box"',
    )

    spp: int = documented(
        attrs.field(default=1000, converter=int, validator=validators.is_positive),
        doc="Number of samples per pixel.",
        type="int",
        default="1000",
    )

    @spp.validator
    def _spp_validator(self, attribute, value):
        if eradiate.mode().is_single_precision and value > 100000:
            warnings.warn(
                f"Measure {self.id} is defined with a sample count greater "
                "than 1e5, but the selected mode is single-precision: results "
                "may be incorrect."
            )

    @property
    @abstractmethod
    def film_resolution(self) -> tuple[int, int]:
        """
        tuple: Getter for film resolution as a (int, int) pair.
        """
        pass

    # --------------------------------------------------------------------------
    #                             Flag-style queries
    # --------------------------------------------------------------------------

    def is_distant(self) -> bool:
        """
        Return ``True`` iff measure records radiometric quantities at infinite
        distance.
        """
        # Default implementation returns False
        return False

    # --------------------------------------------------------------------------
    #                        Kernel dictionary generation
    # --------------------------------------------------------------------------

    @property
    def kernel_type(self) -> str:
        raise NotImplementedError

    @property
    def sensor_id(self) -> str:
        return self.id

    @property
    def template(self) -> dict:
        result = {
            "type": self.kernel_type,
            "id": self.sensor_id,
            "film.type": "hdrfilm",
            "film.width": self.film_resolution[0],
            "film.height": self.film_resolution[1],
            "film.pixel_format": "luminance",
            "film.component_format": "float32",
            "film.rfilter.type": "box",
            "sampler.type": self.sampler,
            "sampler.sample_count": self.spp,
            "medium.type": DictParameter(
                lambda ctx: (
                    "ref"
                    if f"{self.sensor_id}.atmosphere_medium_id" in ctx.kwargs
                    else DictParameter.UNUSED
                ),
            ),
            "medium.id": DictParameter(
                lambda ctx: (
                    ctx.kwargs[f"{self.sensor_id}.atmosphere_medium_id"]
                    if f"{self.sensor_id}.atmosphere_medium_id" in ctx.kwargs
                    else DictParameter.UNUSED
                ),
            ),
        }

        return result

    # --------------------------------------------------------------------------
    #                        Post-processing information
    # --------------------------------------------------------------------------

    @property
    def var(self) -> tuple[str, dict]:
        """
        str, dict: Post-processing variable field name and metadata.
        """
        return "img", dict()



from __future__ import annotations

import attrs
import numpy as np
import pint
import pinttr

from ._core import Spectrum
from ...attrs import define, documented
from ...kernel import DictParameter, KernelSceneParameterFlags, SceneParameter
from ...units import PhysicalQuantity
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


@define(eq=False, slots=False, init=False)
class UniformSpectrum(Spectrum):
    """
    Uniform spectrum [``uniform``] (*i.e.* constant vs wavelength).
    """

    value: pint.Quantity = documented(
        attrs.field(
            converter=lambda x: float(x) if isinstance(x, int) else x,
            repr=lambda x: f"{x:~P}" if isinstance(x, pint.Quantity) else str(x),
            kw_only=True,
        ),
        doc="Uniform spectrum value. If a quantity is passed, units are "
        "checked for consistency. If a unitless value is passed, it is "
        "automatically converted to appropriate configuration default units, "
        "depending on the value of the ``quantity`` field. Integer values "
        "are converted to float. If no quantity is specified, this field must be"
        "a unitless value.",
        type="quantity",
        init_type="quantity or float or int",
    )

    @value.validator
    def _value_validator(self, attribute, value):
        # If a quantity is passed, it must match the quantity field
        if isinstance(value, pint.Quantity):
            if self.quantity is None:
                raise ValueError(
                    f"while validating {attribute.name}, got a Pint quantity, "
                    "incompatible with 'quantity' field set to None; "
                    "please provide a unitless value"
                )
            else:
                expected_units = ucc.get(self.quantity)

                if not pinttr.util.units_compatible(expected_units, value.units):
                    raise pinttr.exceptions.UnitsError(
                        value.units,
                        expected_units,
                        extra_msg=f"while validating {attribute.name}, got units "
                        f"'{value.units}' incompatible with quantity {self.quantity} "
                        f"(expected '{expected_units}')",
                    )

    def __init__(
        self,
        id: str | None = None,
        quantity: PhysicalQuantity | str | None = None,
        *,
        value: int | float | pint.Quantity,
    ):
        # If a quantity is set and a unitless value is passed, it is
        # automatically applied appropriate units
        if quantity is not None and not isinstance(value, pint.Quantity):
            value = pinttr.util.ensure_units(value, ucc.get(quantity))

        self.__attrs_init__(id=id, quantity=quantity, value=value)

    def eval_mono(self, w: pint.Quantity) -> pint.Quantity:
        return np.full_like(w, self.value.m) * self.value.units

    def eval_ckd(self, w: pint.Quantity, g: float) -> pint.Quantity:
        return self.eval_mono(w=w)

    def integral(self, wmin: pint.Quantity, wmax: pint.Quantity) -> pint.Quantity:
        # Convert bounds to unitless values
        wavelength_units = wmin.units
        wmin = wmin.magnitude
        wmax = wmax.m_as(wavelength_units)

        # Compute integral
        integral = self.value * (wmax - wmin)

        # Apply units
        return integral * ureg.dimensionless * wavelength_units

    @property
    def template(self) -> dict:
        # Inherit docstring

        return {
            "type": "uniform",
            "value": DictParameter(
                func=lambda ctx: float(self.eval(ctx.si).m_as(uck.get(self.quantity)))
            ),
        }

    @property
    def params(self) -> dict[str, SceneParameter]:
        # Inherit docstring

        return {
            "value": SceneParameter(
                func=lambda ctx: float(self.eval(ctx.si).m_as(uck.get(self.quantity))),
                flags=KernelSceneParameterFlags.SPECTRAL,
            )
        }



from ._basic import BasicSurface as BasicSurface
from ._central_patch import CentralPatchSurface as CentralPatchSurface
from ._core import Surface as Surface
from ._core import surface_factory as surface_factory
from ._dem import DEMSurface as DEMSurface
from ._dem import mesh_from_dem as mesh_from_dem
from ._dem import triangulate_grid as triangulate_grid



from __future__ import annotations

import typing as t

import mitsuba as mi
import numpy as np
import pint
import pinttr
from pinttr.util import ensure_units

from ._core import ShapeNode
from ..bsdfs import BSDF
from ..core import BoundingBox
from ...attrs import define, documented
from ...contexts import KernelContext
from ...kernel.transform import transform_affine
from ...units import unit_context_config as ucc
from ...units import unit_context_kernel as uck
from ...units import unit_registry as ureg


def _edges_converter(x):
    # Basic unit conversion and array reshaping
    length_units = ucc.get("length")
    x = np.reshape(
        pinttr.util.ensure_units(x, default_units=length_units).m_as(length_units),
        (-1,),
    )

    # Broadcast if relevant
    if len(x) == 1:
        x = np.full((3,), x[0])

    return x * length_units


@define(eq=False, slots=False)
class CuboidShape(ShapeNode):
    """
    Cuboid shape [``cuboid``].

    This shape represents an axis-aligned cuboid parametrized by the length of
    its edges and the coordinates of its central point.

    Notes
    -----
    * If the `to_world` parameter is set, the `center` and `edges` parameters will
      be ignored. The transform will be applied to a cube reaching from
      [-1, -1, -1] to [1, 1, 1].
    """

    center: pint.Quantity = documented(
        pinttr.field(factory=lambda: [0, 0, 0], units=ucc.deferred("length")),
        doc="Coordinates of the centre of the cube. "
        "Unit-enabled field (default: ``ucc['length']``).",
        type="quantity",
        init_type="quantity or array-like, optional",
        default="[0, 0, 0]",
    )

    edges: pint.Quantity = documented(
        pinttr.field(
            factory=lambda: [1, 1, 1],
            converter=_edges_converter,
            units=ucc.deferred("length"),
        ),
        doc="Lengths of the edges of the cuboid. "
        "Unit-enabled field (default: ``ucc['length]``).",
        type="quantity",
        init_type="quantity or array-like",
        default="[1, 1, 1]",
    )

    @property
    def bbox(self) -> BoundingBox:
        # Inherit docstring

        if self.to_world is None:
            return BoundingBox(
                self.center - 0.5 * self.edges, self.center + 0.5 * self.edges
            )
        else:
            new_points = transform_affine(
                self.to_world,
                np.array(
                    [
                        (-1, -1, -1),
                        (1, -1, -1),
                        (-1, 1, -1),
                        (-1, -1, 1),
                        (-1, 1, 1),
                        (1, -1, 1),
                        (1, 1, -1),
                        (1, 1, 1),
                    ]
                ),
            )

            max = new_points.max(axis=0)
            min = new_points.min(axis=0)

            return BoundingBox(min, max)

    def contains(
        self, p: np.typing.ArrayLike, strict: bool = False
    ) -> t.Sequence[bool]:
        """
        Test whether a point lies within the cuboid.

        Parameters
        ----------
        p : quantity or array-like
            An array of shape (3,) (resp. (N, 3)) representing one (resp. N)
            points. If a unitless value is passed, it is interpreted as
            ``ucc['length']``.

        strict : bool
            If ``True``, comparison is done using strict inequalities (<, >).

        Returns
        -------
        result : array of bool or bool
            ``True`` iff ``p`` in within the cuboid.
        """

        length_unit = ucc.get("length")
        p = np.atleast_2d(ensure_units(p, length_unit))

        if self.to_world is None:
            node_min = self.center - 0.5 * self.edges
            node_max = self.center + 0.5 * self.edges

            cmp = (
                np.logical_and(p > node_min, p < node_max)
                if strict
                else np.logical_and(p >= node_min, p <= node_max)
            )
            return np.all(cmp, axis=1)
        else:
            new_points = transform_affine(
                self.to_world,
                np.array(
                    [(0, 0, 0), (-1, -1, -1), (1, -1, -1), (-1, 1, -1), (-1, -1, 1)]
                ),
            )

            p0 = new_points[0]
            p1 = new_points[1]
            p2 = new_points[2]
            p3 = new_points[3]
            p4 = new_points[4]

            v1 = p2 - p1
            l1 = np.linalg.norm(v1)
            n1 = v1 / l1

            v2 = p3 - p1
            l2 = np.linalg.norm(v2)
            n2 = v2 / l2

            v3 = p4 - p1
            l3 = np.linalg.norm(v3)
            n3 = v3 / l3

            p = np.atleast_2d(p.m_as(length_unit))
            vp = p - p0

            projected_1 = np.dot(vp, n1)
            projected_2 = np.dot(vp, n2)
            projected_3 = np.dot(vp, n3)

            l1_half = l1 / 2.0
            l2_half = l2 / 2.0
            l3_half = l3 / 2.0

            if strict:
                result = np.prod(
                    [
                        projected_1 < l1_half,
                        projected_2 < l2_half,
                        projected_3 < l3_half,
                    ],
                    axis=0,
                )
            else:
                result = np.prod(
                    [
                        projected_1 <= l1_half,
                        projected_2 <= l2_half,
                        projected_3 <= l3_half,
                    ],
                    axis=0,
                )
            return result

    def eval_to_world(self, ctx: KernelContext | None = None) -> mi.ScalarTransform4f:
        kwargs = ctx.kwargs.get(self.id, {}) if ctx is not None else {}

        if "to_world" in kwargs:
            return kwargs["to_world"]

        else:
            length_units = uck.get("length")
            converters = {field.name: field.converter for field in self.__attrs_attrs__}

            edges = (
                converters["edges"](kwargs["edges"])
                if "edges" in kwargs
                else self.edges
            )

            center = (
                converters["center"](kwargs["center"])
                if "center" in kwargs
                else self.center
            )

            return mi.ScalarTransform4f.translate(
                center.m_as(length_units)
            ) @ mi.ScalarTransform4f.scale(0.5 * edges.m_as(length_units))

    @property
    def template(self) -> dict:
        length_units = uck.get("length")

        if self.to_world is not None:
            trafo = self.to_world
        else:
            trafo = mi.ScalarTransform4f.translate(
                self.center.m_as(length_units)
            ) @ mi.ScalarTransform4f.scale(0.5 * self.edges.m_as(length_units))

        return {
            "type": "cube",
            "to_world": trafo,
        }

    @classmethod
    def atmosphere(
        cls,
        top: pint.Quantity = 100.0 * ureg.km,
        bottom: pint.Quantity = 0.0 * ureg.km,
        bottom_offset: pint.Quantity = None,
        width: pint.Quantity = 100.0 * ureg.km,
        bsdf: BSDF | None = None,
    ) -> CuboidShape:
        """
        This class method constructor provides a simplified parametrization of
        the cuboid shape better suited for the definition of the atmosphere when
        configuring the one-dimensional model with a plane parallel geometry.

        Parameters
        ----------
        top : quantity, optional, default: 100 km
            Top of atmosphere altitude. If a unitless value is passed, it is
            interpreted as ucc['length'].

        bottom : quantity, optional, default: 0 km
            Ground altitude. If a unitless value is passed, it is interpreted as
            ``ucc['length']``.

        bottom_offset : quantity, optional
            Additional offset by which the cuboid with be extended to avoid an
            exact match of its bottom face and the shape representing the
            surface. If left unset, defaults to a negative offset of  1 % of
            the atmosphere's height, *i.e.*
            :math:`- 0.01 \\times (\\mathtt{top} - \\mathtt{bottom})`.
            If a unitless value is passed, it is interpreted as ``ucc['length']``.

        width : quantity, optional, default: 100 km
            Length of the horizontal edges of the cuboid.
            If a unitless value is passed, it is interpreted as ``ucc['length']``.

        bsdf : BSDF or dict, optional, default: None
            A BSDF specification, forwarded to the main constructor.

        Returns
        -------
        CuboidShape
            A cuboid shape which can be used as the atmosphere in a plane
            parallel geometry.
        """
        top = pinttr.util.ensure_units(top, default_units=ucc.get("length"))
        bottom = pinttr.util.ensure_units(bottom, default_units=ucc.get("length"))
        width = pinttr.util.ensure_units(width, default_units=ucc.get("length"))

        if bottom_offset is None:
            bottom_offset = -0.01 * (top - bottom)
        else:
            bottom_offset = pinttr.util.ensure_units(
                bottom_offset, default_units=ucc.get("length")
            )

        length_units = top.units
        center_z = 0.5 * (top + bottom + bottom_offset).m_as(length_units)
        edge_x = np.squeeze(width.m_as(length_units))
        edge_y = np.squeeze(width.m_as(length_units))
        edge_z = np.squeeze((top - bottom - bottom_offset).m_as(length_units))

        return cls(
            center=[0.0, 0.0, center_z] * length_units,
            edges=[edge_x, edge_y, edge_z] * length_units,
            bsdf=bsdf,
        )



from . import rayleigh as rayleigh
from ._absorption import AbsorptionDatabase as AbsorptionDatabase
from ._absorption import CKDAbsorptionDatabase as CKDAbsorptionDatabase
from ._absorption import ErrorHandlingConfiguration as ErrorHandlingConfiguration
from ._absorption import MonoAbsorptionDatabase as MonoAbsorptionDatabase
from ._array import ArrayRadProfile as ArrayRadProfile
from ._atmosphere import AtmosphereRadProfile as AtmosphereRadProfile
from ._core import RadProfile as RadProfile
from ._core import ZGrid as ZGrid



from __future__ import annotations

import itertools
import typing as t
from abc import ABC, abstractmethod
from functools import singledispatchmethod

import numpy as np
import numpy.typing as npt
import pint
import pinttrs
from pinttrs.util import ensure_units

from .ckd_quad import CKDQuadConfig, CKDQuadPolicy
from .index import CKDSpectralIndex, MonoSpectralIndex, SpectralIndex
from .response import BandSRF, DeltaSRF, SpectralResponseFunction, UniformSRF
from .. import converters
from .._mode import ModeFlag, SubtypeDispatcher
from ..attrs import define, documented
from ..constants import SPECTRAL_RANGE_MAX, SPECTRAL_RANGE_MIN
from ..quad import Quad
from ..radprops import AbsorptionDatabase, CKDAbsorptionDatabase, MonoAbsorptionDatabase
from ..units import unit_context_config as ucc
from ..units import unit_registry as ureg
from ..util.misc import deduplicate_sorted, summary_repr

# ------------------------------------------------------------------------------
#                             Class implementations
# ------------------------------------------------------------------------------


@define
class SpectralGrid(ABC):
    """
    Abstract interface for all spectral grids.
    """

    subtypes = SubtypeDispatcher("SpectralGrid")

    @property
    @abstractmethod
    def wavelengths(self):
        """
        Convenience accessor to characteristic wavelengths of this spectral grid.
        """
        pass

    @staticmethod
    def default() -> SpectralGrid:
        """
        Generate a default spectral grid depending on the active mode.
        """
        cls = SpectralGrid.subtypes.resolve()
        return cls.default()

    @staticmethod
    def arange(
        start: float | pint.Quantity,
        stop: float | pint.Quantity,
        step: float | pint.Quantity,
    ) -> SpectralGrid:
        """
        Generate a spectral grid from equally-spaced wavelengths.

        Parameters
        ----------
        start : quantity or float
            Central wavelength of the first bin. If a unitless value is passed,
            it is interpreted in default wavelength units (usually nm).

        stop : quantity or float
            Wavelength after which bin generation stops. If a unitless value is
            passed, it is interpreted in default wavelength units (usually nm).

        step : quantity or float
            Spectral bin size. If a unitless value is passed, it is interpreted
            in default wavelength units (usually nm).

        Returns
        -------
        SpectralGrid
            Generated spectral grid.
        """
        cls = SpectralGrid.subtypes.resolve()
        return cls.arange(start, stop, step)

    @staticmethod
    def from_absorption_database(abs_db: AbsorptionDatabase) -> SpectralGrid:
        """
        Retrieve the spectral grid from an absorption database. The returned
        type depends on the currently active mode.
        """
        cls = SpectralGrid.subtypes.resolve()
        return cls.from_absorption_database(abs_db)

    def select(self, srf) -> SpectralGrid:
        """
        Select a subset of the spectral grid based on a spectral response
        function.

        Parameters
        ----------
        srf
            A value that is either a :class:`.SpectralResponseFunction` instance
            or convertible to a :class:`.SpectralResponseFunction` by the
            :meth:`.SpectralResponseFunction.convert` method.

        Returns
        -------
        SpectralGrid
            New spectral grid instance covering the extent of the filtering SRF.

        Notes
        -----
        The implementation of this method uses single dispatch based on the type
        of the ``srf`` parameter.
        """
        # This function performs value conversion then calls the _select_impl()
        # dispatching method.
        srf = SpectralResponseFunction.convert(srf)
        return self._select_impl(srf)

    @abstractmethod
    def _select_impl(self, srf: SpectralResponseFunction) -> SpectralGrid:
        pass

    @abstractmethod
    def merge(self, other: SpectralGrid) -> SpectralGrid:
        """
        Merge two spectral grids, applying a boolean "OR" operation.

        Parameters
        ----------
        other : SpectralGrid
            Other spectral, of the same type, to merge with the current one.

        Returns
        -------
        SpectralGrid
            A new spectral grid of the same type that merges the two.
        """
        pass

    @abstractmethod
    def walk_indices(self, **kwargs) -> t.Generator[SpectralIndex, None, None]:
        """
        A generator that yields a sequence of spectral index values.

        Yields
        ------
        .SpectralIndex
            Generated spectral index of a type aligned with the current active
            mode.
        """
        pass


@SpectralGrid.subtypes.register(ModeFlag.SPECTRAL_MODE_MONO)
@define
class MonoSpectralGrid(SpectralGrid):
    """
    A spectral grid consisting of discrete wavelengths, used in monochromatic
    modes.
    """

    _wavelengths: pint.Quantity = documented(
        pinttrs.field(
            units=ucc.deferred("wavelength"),
            converter=[
                pinttrs.converters.to_units(ucc.deferred("wavelength")),
                converters.on_quantity(np.atleast_1d),
                converters.on_quantity(lambda x: x.astype(np.float64)),
                converters.on_quantity(np.unique),
                converters.on_quantity(np.sort),
            ],
            repr=summary_repr,
        ),
        doc="Wavelengths.",
        type="quantity",
        init_type="quantity or array-like or float",
    )

    @property
    def wavelengths(self):
        # Inherit docstring
        return self._wavelengths

    def plot(self, ax, lw=0.5, alpha=1.0):
        ax.vlines(self.wavelengths.m, 0, 1, lw=lw, alpha=alpha)
        return ax

    def _repr_html_(self):
        import base64
        import io

        import matplotlib.pyplot as plt
        import seaborn as sns

        fig, ax = plt.subplots(1, 1, figsize=(6, 1))
        self.plot(ax)
        ax.set_xlabel(f"Wavelength [{self.wavelengths.u:~P}]")
        sns.despine(left=True)
        ax.axes.get_yaxis().set_visible(False)

        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        plt.close(fig)
        img.seek(0)

        return (
            "<img "
            f'src="data:image/png;base64, {base64.b64encode(img.getvalue()).decode("utf-8")}" '
            "/>"
        )

    @staticmethod
    def default() -> MonoSpectralGrid:
        """
        Generate a default monochromatic spectral grid that covers the default
        spectral range with 1 nm spacing.
        """
        return MonoSpectralGrid(
            wavelengths=np.arange(
                SPECTRAL_RANGE_MIN.m_as(ureg.nm),
                SPECTRAL_RANGE_MAX.m_as(ureg.nm) + 0.1,
                1.0,
            )
            * ureg.nm
        )

    @staticmethod
    def arange(
        start: float | pint.Quantity,
        stop: float | pint.Quantity,
        step: float | pint.Quantity,
    ) -> MonoSpectralGrid:
        """
        Generate a spectral grid from equally-spaced wavelengths.

        Parameters
        ----------
        start : quantity or float
            Central wavelength of the first bin. If a unitless value is passed,
            it is interpreted in default wavelength units (usually nm).

        stop : quantity or float
            Wavelength after which bin generation stops. If a unitless value is
            passed, it is interpreted in default wavelength units (usually nm).

        step : quantity or float
            Spectral bin size. If a unitless value is passed, it is interpreted
            in default wavelength units (usually nm).

        Returns
        -------
        MonoSpectralGrid
            Generated spectral grid.
        """
        w_u = ucc.get("wavelength")
        start = ensure_units(start, w_u).m_as(w_u)
        stop = ensure_units(stop, w_u).m_as(w_u)
        step = ensure_units(step, w_u).m_as(w_u)
        return MonoSpectralGrid(wavelengths=np.arange(start, stop, step) * w_u)

    @classmethod
    def from_absorption_database(cls, abs_db: MonoAbsorptionDatabase):
        """
        Retrieve the spectral grid from a monochromatic absorption database.
        """
        if not isinstance(abs_db, MonoAbsorptionDatabase):
            raise TypeError

        w = abs_db.spectral_coverage.index.get_level_values(level=1).values * ureg.nm
        return cls(wavelengths=w)

    @singledispatchmethod
    def _select_impl(self, srf: SpectralResponseFunction) -> MonoSpectralGrid:
        # Inherit docstring
        raise NotImplementedError(f"unsupported data type '{type(srf)}'")

    @_select_impl.register
    def _(self, srf: DeltaSRF):
        # Pass SRF wavelengths through
        return MonoSpectralGrid(wavelengths=srf.wavelengths)

    @_select_impl.register
    def _(self, srf: UniformSRF):
        w_m = self.wavelengths.m
        w_u = self.wavelengths.u
        wmin_m, wmax_m = srf.wmin.m_as(w_u), srf.wmax.m_as(w_u)

        w_selected_m = w_m[(w_m >= wmin_m) & (w_m <= wmax_m)]
        return MonoSpectralGrid(wavelengths=w_selected_m * w_u)

    @_select_impl.register
    def _(self, srf: BandSRF):
        # Select all wavelengths for which the SRF evaluates to a nonzero value
        values = srf.eval(self.wavelengths)
        w_selected = self.wavelengths[values.m > 0.0]
        return MonoSpectralGrid(wavelengths=w_selected)

    def merge(self, other: MonoSpectralGrid) -> MonoSpectralGrid:
        # Inherit docstring

        # Collect all wavelengths
        w_u = ucc.get("wavelength")
        w_m = np.sort(
            np.concatenate((self.wavelengths.m_as(w_u), other.wavelengths.m_as(w_u)))
        )

        # Remove duplicates
        w_m = np.unique(w_m)

        return MonoSpectralGrid(wavelengths=w_m * w_u)

    def walk_indices(self) -> t.Generator[MonoSpectralIndex, None, None]:
        # Inherit docstring
        for w in self.wavelengths:
            yield MonoSpectralIndex(w=w)


@SpectralGrid.subtypes.register(ModeFlag.SPECTRAL_MODE_CKD)
@define(init=False)
class CKDSpectralGrid(SpectralGrid):
    """
    A spectral grid that splits the spectral dimensions into bins characterized
    by their bounds.

    Parameters
    ----------
    fix_bounds : {"keep_min", "keep_max"} or False, default: "keep_min"
        Unless told no to, the constructor will detect lower and upper bound
        values close to each other within a tolerance and flag them as matching.
        If this parameter is set to ``"keep_min"`` or ``"keep_max"``, the
        constructor make sure that matching bounds effectively match exactly.
        If it is set to ``"raise"``, it will raise an exception. If it is set to
        ``"ignore"``, no action will be taken. The tolerance is controlled by
        the ``epsilon`` parameter.

    epsilon : float, default: 1e-6
        Absolute tolerance for matching bound detection.

    Raises
    ------
    ValueError
        If matching bound misalignment is detected and ``fix_bounds`` is set to
        ``"raise"``.
    """

    wmins: pint.Quantity = documented(
        pinttrs.field(units=ucc.deferred("wavelength"), repr=summary_repr),
        doc="Lower bound of all bins. Unitless values are interpreted as default "
        "wavelength config units.",
        type="quantity",
        init_type="quantity or array-like",
    )

    wmaxs: pint.Quantity = documented(
        pinttrs.field(units=ucc.deferred("wavelength"), repr=summary_repr),
        doc="Upper bound of all bins. Unitless values are interpreted as default "
        "wavelength config units.",
        type="quantity",
        init_type="quantity or array-like",
    )

    wcenters: pint.Quantity = documented(
        pinttrs.field(units=ucc.deferred("wavelength"), repr=summary_repr),
        doc="Central wavelength of all bins. Unitless values are interpreted as "
        "default wavelength config units. "
        "If unset, bin centers are computed automatically from bin bounds. "
        "Bin centers are allowed to be different from the middle of the bin "
        "interval and, when the grid is tied to a database, are expected to "
        "match the values of the wavelength coordinate in the database. However, "
        "this is considered as a workaround to deal with poorly indexed databases, "
        "and users should try to set central wavelengths to the middle of spectral "
        "bins.",
        type="quantity, optional",
        init_type="quantity or array-like",
    )

    def __init__(
        self,
        wmins: npt.ArrayLike,
        wmaxs: npt.ArrayLike,
        wcenters: npt.ArrayLike | None = None,
        fix_bounds: t.Literal["keep_min", "keep_max", "raise", "ignore"] = "keep_min",
        epsilon: float = 1e-6,
    ):
        # Ensure consistent units and appropriate dtype
        w_u = ucc.get("wavelength")
        wmins_m = ensure_units(wmins, w_u).m_as(w_u).astype(np.float64)
        wmaxs_m = ensure_units(wmaxs, w_u).m_as(w_u).astype(np.float64)

        # Detect bound mismatch
        diff_bounds = wmaxs_m[:-1] - wmins_m[1:]
        fix_locations = (diff_bounds > 0.0) & (diff_bounds <= epsilon)
        if np.any(fix_locations):
            if fix_bounds == "keep_max":
                wmins_m[1:] = np.where(fix_locations, wmaxs_m[:-1], wmins_m[1:])
            elif fix_bounds == "keep_min":
                wmaxs_m[:-1] = np.where(fix_locations, wmins_m[1:], wmaxs_m[:-1])
            elif fix_bounds == "raise":
                raise ValueError(
                    "while constructing CKDSpectralGrid: bin bound mismatch "
                    f"(min: {wmins_m[1:][fix_locations]}; max: {wmaxs_m[:-1][fix_locations]}"
                )
            elif fix_bounds == "ignore":
                pass
            else:
                raise ValueError(f'unknown bound fixing policy "{fix_bounds}"')

        # Define bin centers if necessary
        if wcenters is None:
            wcenters = 0.5 * (wmins_m + wmaxs_m) * w_u

        # Initialize the object
        self.__attrs_init__(wmins_m * w_u, wmaxs_m * w_u, wcenters)

    @property
    def wavelengths(self):
        # Inherit docstring
        return self.wcenters

    def plot(self, ax, alpha=0.5):
        import seaborn as sns
        from cycler import cycler

        w_u = ucc.get("wavelength")
        color_cycle = cycler(color=sns.color_palette())

        for wmin, wmax, wcenter, color in zip(
            self.wmins.m_as(w_u),
            self.wmaxs.m_as(w_u),
            self.wcenters.m_as(w_u),
            itertools.cycle(color_cycle),
        ):
            c = color["color"]
            ax.fill_between(
                [wmin, wmax], 0, 1, color=c, alpha=alpha, lw=0.5, ls=(0, (5, 5))
            )
            ax.vlines(wcenter, 0, 1, color=c, lw=0.5)

        return ax

    def _repr_html_(self):
        import base64
        import io

        import matplotlib.pyplot as plt
        import seaborn as sns

        w_u = ucc.get("wavelength")

        fig, ax = plt.subplots(1, 1, figsize=(6, 1))
        self.plot(ax)
        ax.set_xlabel(f"Wavelength [{w_u:~P}]")
        sns.despine(left=True)
        ax.axes.get_yaxis().set_visible(False)

        img = io.BytesIO()
        fig.savefig(img, format="png", bbox_inches="tight")
        plt.close(fig)
        img.seek(0)

        return (
            "<img "
            f'src="data:image/png;base64, {base64.b64encode(img.getvalue()).decode("utf-8")}" '
            "/>"
        )

    @staticmethod
    def default() -> CKDSpectralGrid:
        """
        Generate a default CKD spectral that covers the default spectral range
        with 10 nm spacing.
        """
        return CKDSpectralGrid.arange(
            start=SPECTRAL_RANGE_MIN.m_as(ureg.nm),
            stop=SPECTRAL_RANGE_MAX.m_as(ureg.nm) + 1.0,
            step=10.0,
        )

    @staticmethod
    def arange(
        start: float | pint.Quantity,
        stop: float | pint.Quantity,
        step: float | pint.Quantity,
    ) -> CKDSpectralGrid:
        """
        Generate a CKD spectral grid with equally-sized bins.

        Parameters
        ----------
        start : quantity or float
            Central wavelength of the first bin. If a unitless value is passed,
            it is interpreted in default wavelength units (usually nm).

        stop : quantity or float
            Wavelength after which bin generation stops. If a unitless value is
            passed, it is interpreted in default wavelength units (usually nm).

        step : quantity or float
            Spectral bin size. If a unitless value is passed, it is interpreted
            in default wavelength units (usually nm).

        Returns
        -------
        CKDSpectralGrid
            Generated CKD spectral grid.
        """
        w_u = ucc.get("wavelength")
        start_m = ensure_units(start, w_u).m_as(w_u)
        stop_m = ensure_units(stop, w_u).m_as(w_u)
        width_m = ensure_units(step, w_u).m_as(w_u)

        wcenters_m = np.arange(start_m, stop_m, width_m)
        wmins_m = wcenters_m - 0.5 * width_m
        wmaxs_m = wcenters_m + 0.5 * width_m

        return CKDSpectralGrid(wmins_m * w_u, wmaxs_m * w_u, wcenters_m * w_u)

    @classmethod
    def from_nodes(cls, wnodes: npt.ArrayLike) -> CKDSpectralGrid:
        wmins = wnodes[:-1]
        wmaxs = wnodes[1:]
        return cls(wmins=wmins, wmaxs=wmaxs)

    @classmethod
    def from_absorption_database(cls, abs_db: CKDAbsorptionDatabase) -> CKDSpectralGrid:
        """
        Retrieve the spectral grid from a CKD absorption database.

        Parameters
        ----------
        abs_db : .CKDAbsorptionDatabase
        """
        if not isinstance(abs_db, CKDAbsorptionDatabase):
            raise TypeError(
                "CKD spectral grid can only be derived from a "
                f"CKDAbsorptionDatabase instance, got a {type(abs_db).__name__}"
            )

        wmins = abs_db.spectral_coverage["wbound_lower [nm]"].values * ureg.nm
        wmaxs = abs_db.spectral_coverage["wbound_upper [nm]"].values * ureg.nm
        wcenters = abs_db.spectral_coverage.index.get_level_values(1).values * ureg.nm
        return cls(wmins, wmaxs, wcenters)

    @singledispatchmethod
    def _select_impl(self, srf: SpectralResponseFunction) -> CKDSpectralGrid:
        # Inherit docstring
        raise NotImplementedError(f"unsupported data type '{type(srf)}'")

    @_select_impl.register
    def _(self, srf: DeltaSRF):
        w_u = srf.wavelengths.u
        w_m = srf.wavelengths.m
        wmins_m = self.wmins.m_as(w_u)
        wmaxs_m = self.wmaxs.m_as(w_u)

        selmin = np.searchsorted(wmins_m, w_m)
        selmax = np.searchsorted(wmaxs_m, w_m) + 1
        hit = selmin == selmax  # Mask where w_m values which triggered a bin hit

        # Map w values to selected bin (index -999 means not selected)
        bin_index = np.where(hit, selmin - 1, np.full_like(w_m, -999)).astype(np.int64)

        # Get selected bins only
        selected = np.unique(bin_index)  # mask removes -999 value
        selected = selected[selected >= 0]

        return CKDSpectralGrid(wmins=self.wmins[selected], wmaxs=self.wmaxs[selected])

    @_select_impl.register
    def _(self, srf: UniformSRF):
        selected = (self.wmaxs > srf.wmin) & (self.wmins < srf.wmax)
        return CKDSpectralGrid(wmins=self.wmins[selected], wmaxs=self.wmaxs[selected])

    @_select_impl.register
    def _(self, srf: BandSRF):
        w_u = self.wmins.u
        wmins_m = self.wmins.m_as(w_u)
        wmaxs_m = self.wmaxs.m_as(w_u)

        # Build spectral mesh used to interpolate
        w_m = np.unique(np.concatenate((wmins_m, wmaxs_m)))
        # Note the handling of numeric precision-induced min and max bin bound
        # mismatch was removed from the previous implementation because
        # consistency is enforced upon initialization

        # Detect spectral bins on which the SRF takes nonzero values
        cumsum = np.concatenate(([0], srf.integrate_cumulative(w_m * w_u).m_as(w_u)))
        selected = cumsum[:-1] != cumsum[1:]

        # Build a new spectral grid that only contains selected bins
        return CKDSpectralGrid(self.wmins[selected], self.wmaxs[selected])

    def merge(self, other: CKDSpectralGrid) -> CKDSpectralGrid:
        # Inherit docstring

        # Collect spectral bin information
        w_u = ucc.get("wavelength")
        wmins = np.concatenate((self.wmins.m_as(w_u), other.wmins.m_as(w_u)))
        wmaxs = np.concatenate((self.wmaxs.m_as(w_u), other.wmaxs.m_as(w_u)))
        wcenters = np.concatenate((self.wcenters.m_as(w_u), other.wcenters.m_as(w_u)))

        # Sort bins and remove duplicates
        # TODO: Vectorize this for best performance, this is quick and dirty
        w_m = sorted(
            np.stack((wmins, wmaxs, wcenters)).T.tolist(),
            key=lambda x: (x[0], x[1], x[2]),
        )
        w_m = np.array(deduplicate_sorted(w_m))

        return CKDSpectralGrid(
            wmins=w_m[:, 0] * w_u, wmaxs=w_m[:, 1] * w_u, wcenters=w_m[:, 2] * w_u
        )

    def walk_quads(
        self,
        ckd_quad_config: CKDQuadConfig,
        abs_db: CKDAbsorptionDatabase | None = None,
    ) -> t.Generator[tuple[pint.Quantity, Quad]]:
        """
        Walk the spectral grid and retrieve, based on a quadrature configuration
        and, if necessary, an absorption database, the spectral quadrature for
        each spectral bin.

        Parameters
        ----------
        ckd_quad_config : .CKDQuadConfig
            CKD quadrature configuration.

        abs_db : .CKDAbsorptionDatabase, optional
            Molecular absorption database used to build quadrature rules for
            each spectral bin. This parameter is required only if an adaptive
            quadrature generation policy is used, otherwise it is ignored.

        Yields
        ------
        quad : .Quad
            Quadrature rule for the current spectral bin.

        w : quantity
            Wavelength of the current spectral bin.
        """

        # Check parameter consistency
        if ckd_quad_config.policy is not CKDQuadPolicy.FIXED and abs_db is None:
            raise ValueError(
                "while attempting CKD spectral grid walk with policy "
                f"{ckd_quad_config.policy}: `abs_db` must be set (got None)"
            )

        # Walk the spectral grid and get the quadrature for each bin
        for w in self.wcenters:
            yield w, ckd_quad_config.get_quad(abs_db, wcenter=w)

    def walk_indices(
        self,
        ckd_quad_config: CKDQuadConfig,
        abs_db: CKDAbsorptionDatabase | None = None,
    ) -> t.Generator[CKDSpectralIndex]:
        """
        Walk the spectral grid and retrieve, based on a quadrature configuration
        and, if necessary, an absorption database, the sequence of spectral
        indexes driving the spectral loop.

        Parameters
        ----------
        ckd_quad_config : .CKDQuadConfig
            CKD quadrature configuration.

        abs_db : .CKDAbsorptionDatabase, optional
            Molecular absorption database used to build quadrature rules for
            each spectral bin. This parameter is required only if an adaptive
            quadrature generation policy is used, otherwise it is ignored.

        Yields
        ------
        si : .CKDSpectralIndex
            Generated spectral index.
        """

        # Walk the spectral dimension
        for w, quad in self.walk_quads(ckd_quad_config, abs_db):
            for g in quad.eval_nodes([0, 1]):
                yield CKDSpectralIndex(w=w, g=g)



"""
Molecular atmospheres.
"""

from __future__ import annotations

import attrs
import joseki
import numpy as np
import pint
import xarray as xr

import eradiate

from ._core import AbstractHeterogeneousAtmosphere
from ..core import traverse
from ..phase import PhaseFunction, RayleighPhaseFunction
from ...attrs import define, documented
from ...contexts import KernelContext
from ...converters import convert_thermoprops
from ...exceptions import UnsupportedModeError
from ...radprops import (
    AbsorptionDatabase,
    AtmosphereRadProfile,
    ErrorHandlingConfiguration,
    RadProfile,
    ZGrid,
)
from ...spectral.index import SpectralIndex
from ...units import unit_registry as ureg
from ...util.misc import summary_repr


def _default_absorption_data():
    if eradiate.mode().is_mono:
        return "komodo"
    elif eradiate.mode().is_ckd:
        return "monotropa"
    else:
        raise UnsupportedModeError(unsupported=["mono", "ckd"])


@define(eq=False, slots=False)
class MolecularAtmosphere(AbstractHeterogeneousAtmosphere):
    """
    Molecular atmosphere scene element [``molecular``].

    See Also
    --------
    :class:`~eradiate.scenes.atmosphere.ParticleLayer`,
    :class:`~eradiate.scenes.atmosphere.HeterogeneousAtmosphere`.

    Notes
    -----
    This is commonly referred to as a clear-sky
    atmosphere, namely the atmosphere is free of clouds, aerosols or any other
    type of liquid or solid particles in suspension.

    It describes a gaseous mixture (air) whose thermophysical properties,
    namely pressure, temperature and constituent mole fractions, are allowed to
    vary with altitude (see the ``thermoprops`` attribute).

    The corresponding radiative properties are computed with this
    thermophysical profile as input.

    Special care must be taken that the absorption data is able to accomodate
    the specified thermophysical profile, especially along the constituent mole
    fractions axes (see also ``error_handler_config``).
    Note that the absoprtion data is able to support one spectral mode at a
    time. As a result, the Eradiate mode must be selected before instantiating
    this class, and the relevant absorption data must be provided.

    The Rayleigh scattering theory is used to compute the air volume
    scattering coefficient.

    The scattering phase function defaults to the Rayleigh scattering phase
    function but can be set to other values.
    """

    _absorption_data: AbsorptionDatabase = documented(
        attrs.field(
            kw_only=True,
            factory=AbsorptionDatabase.default,
            converter=AbsorptionDatabase.convert,
            validator=attrs.validators.instance_of(AbsorptionDatabase),
        ),
        doc="Absorption coefficient data. The passed value is pre-processed by "
        ":meth:`.AbsorptionDatabase.convert`.",
        type="AbsorptionDatabase",
        init_type="str or path-like or dict or .AbsorptionDatabase",
        default=":meth:`AbsorptionDatabase.default() <.AbsorptionDatabase.default>`",
    )

    _thermoprops: xr.Dataset | None = documented(
        attrs.field(
            kw_only=True,
            factory=lambda: joseki.make(
                identifier="afgl_1986-us_standard",
                z=np.linspace(0.0, 120.0, 121) * ureg.km,
                additional_molecules=False,
            ),
            converter=attrs.converters.optional(convert_thermoprops),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(xr.Dataset)
            ),
            repr=summary_repr,
        ),
        doc="Thermophysical property dataset. If a path is passed, Eradiate will "
        "look it up and load it. If a dictionary is passed, it will be passed "
        "as keyword argument to ``joseki.make()``. The default is "
        '``joseki.make(identifier="afgl_1986-us_standard",  z=np.linspace(0.0, 120.0, 121) * ureg.km)``. '
        "See the `Joseki documentation <https://joseki.readthedocs.io/en/latest/reference/joseki/index.html#joseki.make>`__ "
        "for details.",
        type="Dataset or None",
        init_type="Dataset or path-like or dict or None",
    )

    has_absorption: bool = documented(
        attrs.field(kw_only=True, default=True, converter=bool),
        doc="Absorption switch. If ``True``, the absorption coefficient is "
        "computed. Else, the absorption coefficient is set to zero.",
        type="bool",
        default="True",
    )

    has_scattering: bool = documented(
        attrs.field(kw_only=True, default=True, converter=bool),
        doc="Scattering switch. If ``True``, the scattering coefficient is "
        "computed. Else, the scattering coefficient is set to zero.",
        type="bool",
        default="True",
    )

    rayleigh_depolarization: np.ndarray | str = documented(
        attrs.field(
            converter=lambda x: x
            if isinstance(x, str)
            else np.array(x, dtype=np.float64),
            kw_only=True,
            factory=lambda: np.array(0.0),
        ),
        type='ndarray or {"bates", "bodhaine"}',
        doc="Depolarization factor of the rayleigh phase function. "
        "``str`` will be interpreted as the name of the function used to "
        "calculate the depolarization factor from atmospheric properties. "
        "A ``ndarray`` will be interpreted as a description of the depolarization "
        "factor at different levels of the atmosphere. Must be shaped (N,) with "
        "N the number of layers.",
    )

    @has_absorption.validator
    @has_scattering.validator
    def _switch_validator(self, attribute, value):
        if not self.has_absorption and not self.has_scattering:
            raise ValueError(
                f"while validating {attribute.name}: at least one of "
                "'has_absorption' and 'has_scattering' must be True"
            )

    _radprops_profile: RadProfile | None = documented(
        attrs.field(
            kw_only=True,
            default=None,
            validator=attrs.validators.optional(
                attrs.validators.instance_of(RadProfile)
            ),
        ),
        doc="Radiative property profile. "
        "If ``thermoprops`` is not ``None``, this field is automatically "
        "overridden with an :class:`.AtmosphereRadProfile` during initialization. "
        "Note that at least ``thermoprops`` or ``radprops_profile`` should not "
        "be ``None``.",
        type=".RadProfile or None",
        init_type=".RadProfile or None",
        default="None",
    )

    error_handler_config: ErrorHandlingConfiguration | None = documented(
        attrs.field(
            kw_only=True,
            default=None,
            converter=ErrorHandlingConfiguration.convert,
            validator=attrs.validators.optional(
                attrs.validators.instance_of(ErrorHandlingConfiguration)
            ),
        ),
        doc="Error handler configuration for absorption data interpolation. If "
        "unset, the global configuration specified in the "
        "``absorption_database.error_handling`` section is used. ",
        type=".ErrorHandlingConfiguration or None",
        init_type="dict or .ErrorHandlingConfiguration, optional",
    )

    def update(self) -> None:
        # Inherit docstring
        self.phase.id = self.phase_id

        if self.thermoprops is not None:
            self._radprops_profile = AtmosphereRadProfile(
                thermoprops=self.thermoprops,
                has_scattering=self.has_scattering,
                has_absorption=self.has_absorption,
                absorption_data=self.absorption_data,
                rayleigh_depolarization=self.rayleigh_depolarization,
            )

    # --------------------------------------------------------------------------
    #              Spatial extension and thermophysical properties
    # --------------------------------------------------------------------------

    @property
    def thermoprops(self) -> xr.Dataset:
        # Inherit docstring
        return self._thermoprops

    def eval_mfp(self, ctx: KernelContext) -> pint.Quantity:
        # Inherit docstring
        min_sigma_s = self.radprops_profile.eval_sigma_s(ctx.si).min()
        return np.divide(
            1.0,
            min_sigma_s,
            where=min_sigma_s != 0.0,
            out=np.array([np.inf]),
        )

    # --------------------------------------------------------------------------
    #                             Radiative properties
    # --------------------------------------------------------------------------

    @property
    def absorption_data(self) -> AbsorptionDatabase:
        # Inherit docstring
        return self._absorption_data

    @property
    def phase(self) -> PhaseFunction:
        # Inherit docstring

        def eval_depolarization_factor(si: SpectralIndex) -> np.ndarray:
            return self.eval_depolarization_factor(si).m_as("dimensionless")

        # pass callable for depolarization to phase function for InitParams and UpdateParams.
        return RayleighPhaseFunction(
            depolarization=eval_depolarization_factor, geometry=self.geometry
        )

    @property
    def radprops_profile(self) -> RadProfile:
        # Inherit docstring
        return self._radprops_profile

    def eval_albedo(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_albedo(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_t(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_t(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_a(
        self, si: SpectralIndex, zgrid: ZGrid | None = None, **kwargs
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_a(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_sigma_s(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        # Inherit docstring
        return self.radprops_profile.eval_sigma_s(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    def eval_depolarization_factor(
        self, si: SpectralIndex, zgrid: ZGrid | None = None
    ) -> pint.Quantity:
        return self.radprops_profile.eval_depolarization_factor(
            si,
            zgrid=self.geometry.zgrid if zgrid is None else zgrid,
        )

    # --------------------------------------------------------------------------
    #                             Kernel dictionary
    # --------------------------------------------------------------------------

    @property
    def _template_phase(self) -> dict:
        # Inherit docstring
        result, _ = traverse(self.phase)
        return result.data

    @property
    def _params_phase(self) -> dict:
        # Inherit docstring
        _, result = traverse(self.phase)
        return result.data



from __future__ import annotations

import attrs

from ._core import Illumination
from ..core import NodeSceneElement
from ..spectra import Spectrum, spectrum_factory
from ...attrs import define, documented
from ...validators import has_quantity


@define(eq=False, slots=False)
class ConstantIllumination(Illumination):
    """
    Constant illumination scene element [``constant``].
    """

    radiance: Spectrum = documented(
        attrs.field(
            default=1.0,
            converter=spectrum_factory.converter("radiance"),
            validator=[
                attrs.validators.instance_of(Spectrum),
                has_quantity("radiance"),
            ],
        ),
        doc="Emitted radiance spectrum. Must be a radiance spectrum "
        "(in W/m²/sr/nm or compatible units).",
        type=":class:`~eradiate.scenes.spectra.Spectrum`",
        init_type=":class:`~eradiate.scenes.spectra.Spectrum` or dict or float",
        default="1.0 ucc[radiance]",
    )

    @property
    def template(self) -> dict:
        return {"type": "constant"}

    @property
    def objects(self) -> dict[str, NodeSceneElement]:
        return {"radiance": self.radiance}



import enum
import logging
from typing import Annotated, List, Optional

import typer

from eradiate import asset_manager, fresolver

from ._console import message, section

app = typer.Typer()

logger = logging.getLogger(__name__)


class ListWhat(str, enum.Enum):
    resources = "resources"
    aliases = "aliases"
    all = "all"


class ClearWhat(str, enum.Enum):
    cached = "cached"
    unpacked = "unpacked"
    installed = "installed"
    all = "all"


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """
    Display the asset manager and file resolver configurations.
    Use subcommands for other data management tasks.
    """
    if ctx.invoked_subcommand is None:
        asset_manager_info = asset_manager.info()
        section("Asset manager")
        message(f"• Remote storage URL: {asset_manager_info['remote_url']}")
        message(
            f"• Asset cache location [{asset_manager_info['cache_size']:.3g~P}]: "
            f"{asset_manager_info['cache_dir']}"
        )
        message(
            f"• Unpacked asset location [{asset_manager_info['unpack_size']:.3g~P}]: "
            f"{asset_manager_info['unpack_dir']}"
        )
        message(f"• Installation location: {asset_manager_info['install_dir']}")

        fresolver_info = fresolver.info()
        section("File resolver")
        for path in fresolver_info["paths"]:
            message(f"• {path}")


@app.command()
def update():
    """
    Download the data registry manifest from the remote data location.
    """
    asset_manager.update(download=True)


@app.command()
def list(
    what: Annotated[
        ListWhat,
        typer.Option(help="A keyword that specifies what to clear."),
    ] = ListWhat.resources,
    aliases: Annotated[bool, typer.Option(help="Alias to --what aliases.")] = False,
    all: Annotated[bool, typer.Option(help="Alias to --what all.")] = False,
):
    """
    List all packages referenced by the manifest and their current state
    (cached, unpacked, installed).
    """
    if aliases:
        what = ListWhat.aliases
    if all:
        what = ListWhat.all

    asset_manager.update()
    asset_manager.list(what=what)


@app.command()
def download(
    resource_ids: Annotated[
        List[str], typer.Argument(help="One or multiple resource IDs or aliases.")
    ],
    unpack: Annotated[bool, typer.Option(help="Unpack downloaded archives.")] = True,
):
    """
    Download a resource from remote storage to the cache directory.
    """
    asset_manager.download(resource_ids, unpack=unpack, progressbar=True)


@app.command()
def install(
    resource_ids: Annotated[
        List[str], typer.Argument(help="One or multiple resource IDs or aliases.")
    ],
):
    """
    Install a resource. If the data is not already cached locally, it is
    downloaded from remove storage.
    """
    asset_manager.install(resource_ids)


@app.command()
def remove(
    resource_ids: Annotated[
        List[str], typer.Argument(help="One or multiple resource IDs or aliases.")
    ],
):
    """
    Uninstall a resource.
    """
    asset_manager.remove(resource_ids)


@app.command()
def clear(
    resource_ids: Annotated[
        Optional[List[str]],
        typer.Argument(
            help="Resource(s) for which to clear data. If unset, all data will be wiped."
        ),
    ] = None,
    what: Annotated[
        ClearWhat,
        typer.Option(help="A keyword that specifies what to clear."),
    ] = ClearWhat.cached,
    all: Annotated[bool, typer.Option(help="Alias to --what all.")] = False,
    unpacked: Annotated[bool, typer.Option(help="Alias to --what unpacked.")] = False,
    installed: Annotated[bool, typer.Option(help="Alias to --what installed.")] = False,
):
    """
    Delete data.
    """
    if unpacked:
        what = ClearWhat.unpacked
    if installed:
        what = ClearWhat.installed
    if all:
        what = ClearWhat.all

    asset_manager.clear(resource_ids, what=what)



from __future__ import annotations

__all__ = ["dataarray_to_rgb", "film_to_angular"]

import typing as t

import numpy as np
import xarray as xr

from .. import frame
from ..config import settings
from ..warp import uniform_hemisphere_to_square


def film_to_angular(
    da: xr.DataArray,
    theta: np.typing.ArrayLike,
    phi: np.typing.ArrayLike,
    x_label: str = "x",
    y_label: str = "y",
    theta_label: str = "theta",
    phi_label: str = "phi",
    azimuth_convention: frame.AzimuthConvention | str | None = None,
) -> xr.DataArray:
    """
    Interpolate a hemispherical film data array on an angular grid.

    Parameters
    ----------
    da : DataArray
        Data array with film coordinate pixels to interpolate on an angular grid.

    theta : array-like
        List of zenith angle values.

    phi : array-like
        List of azimuth angle values.

    x_label : str, default: "x"
        Label for the width pixel coordinate.

    y_label : str, default: "y"
        Label for the height pixel coordinate.

    theta_label : str, default: "theta"
        Label for the zenith angle coordinate.

    phi_label : str, default: "phi"
        Label for the azimuth angle coordinate.

    azimuth_convention : .AzimuthConvention or str, optional
        Azimuth convention used in the produced data array. If unset the default
        active convention is used.

    Returns
    -------
    DataArray
        Data array interpolated on the specified angular grid.
    """
    # TODO: Double check where this function is used and if the azimuth
    #       transformation is correctly applied.

    # Define azimuth convention
    if azimuth_convention is None:
        azimuth_convention = settings.azimuth_convention
    else:
        azimuth_convention = frame.AzimuthConvention.convert(azimuth_convention)

    # Map angular grid points to (x, y) space and interpolate film data accordingly
    das = []
    for ph in phi:
        xs = np.empty_like(theta)
        ys = np.empty_like(theta)

        angles = np.array([[th, ph] for th in theta.ravel()])
        directions = frame.angles_to_direction(angles)
        film_coords = uniform_hemisphere_to_square(directions)
        xs.ravel()[:] = film_coords[:, 0]
        ys.ravel()[:] = film_coords[:, 1]

        # Interpolate at target angular coordinates
        x = xr.DataArray(xs, dims=theta_label)
        y = xr.DataArray(ys, dims=theta_label)
        interpolated = da.interp(**{x_label: x, y_label: y})

        # Introduce theta as dimension coordinate
        interpolated = interpolated.assign_coords({theta_label: theta})

        # Drop non-dimension coordinates
        interpolated = interpolated.drop_vars(
            [name for name in da.coords if name not in interpolated.dims]
        )

        # Add phi as dimension coordinate (used later when concatenating)
        interpolated = interpolated.expand_dims(
            {
                phi_label: [
                    frame.transform_azimuth(ph, to_convention=azimuth_convention)
                ]
            },
            axis=-1,
        )
        das.append(interpolated)

    # Concatenate all azimuth slices
    result = xr.concat(das, dim=phi_label)
    return result


def dataarray_to_rgb(
    da: xr.DataArray,
    channels: t.Sequence[tuple[str, t.Any]],
    normalize: bool = True,
    gamma_correction: bool = True,
) -> np.ndarray:
    """
    Compose an RGB image from radiance data.

    Parameters
    ----------
    da : DataArray
        The data array from which radiance data will be taken. It has to be such
        that data, when selected on a spectral axis, is 2-dimensional.

    channels : sequence of tuples
        Three (coordinate label, coordinate value) pairs used to select the data
        used to compose the image. Channels are ordered as follows: (R, G, B).
        For instance, to select wavelengths (dimension ``"w"``) at 440 (blue),
        550 (green) and 660 (red) nm, use
        ``channels=[("w", 660), ("w", 550), ("w", 440)]``.

    normalize : bool, default: True
        If ``True``, the data will be normalized by its maximum value.

    gamma_correction : bool, default: True
        If ``True``, apply a gamma operator to the data.

    Returns
    -------
    ndarray
        An RGB image which can be displayed using
        :func:`matplotlib.pyplot.imshow`.

    Warnings
    --------
    The image processing pipeline implemented by this function is rudimentary.
    It only applies, if instructed, a gamma operator. For more advanced tone
    mapping operations, use this function with ``normalize=False`` and
    ``gamma_correction=False``, then apply your own post-processing to the
    resulting (N, M, 3)-shaped array.
    """
    if len(channels) != 3:
        raise ValueError("channel list must have 3 elements (R, G, B)")

    # Collect data
    result = []

    for coord, value in channels:
        x = da.sel(**{coord: value}).squeeze().values
        if x.ndim != 2:
            raise ValueError("only 2D arrays can be assembled into an RGB image")
        result.append(np.expand_dims(x, axis=-1))

    result = np.concatenate(result, axis=2)

    # Normalize to [0, 1] interval
    if normalize:
        result /= np.nanmax(result)

    # Apply gamma correction
    if gamma_correction:
        result **= 1.0 / 2.2

    return result



from __future__ import annotations

import typing as t

import attrs
import mitsuba as mi
import numpy as np

import eradiate

from ._core import PhaseFunction
from ..geometry import PlaneParallelGeometry, SceneGeometry, SphericalShellGeometry
from ...attrs import define, documented
from ...contexts import KernelContext
from ...kernel import DictParameter, KernelSceneParameterFlags, SceneParameter
from ...spectral.index import SpectralIndex


@define(eq=False, slots=False)
class RayleighPhaseFunction(PhaseFunction):
    """
    Rayleigh phase function [``rayleigh``].

    The Rayleigh phase function models scattering by particles with a
    characteristic size much smaller than the considered radiation wavelength.
    It is typically used to represent scattering by gas molecules in the
    atmosphere.
    """

    depolarization: np.ndarray | t.Callable[[KernelContext], np.ndarray] = documented(
        attrs.field(
            converter=lambda x: x if callable(x) else np.array(x, dtype=np.float64),
            kw_only=True,
            factory=lambda: np.array([0.0]),
        ),
        doc="The ratio of intensities parallel and perpendicular to the "
        "plane of scattering for light scattered at 90 deg. Only relevant "
        "when using a polarization mode.",
        type="ndarray, callable or None",
        init_type="array-like, callable, or None",
    )

    geometry: SceneGeometry | None = documented(
        attrs.field(
            default=None,
            converter=attrs.converters.optional(SceneGeometry.convert),
            validator=attrs.validators.optional(
                attrs.validators.instance_of(SceneGeometry)
            ),
        ),
        doc="Parameters defining the basic geometry of the scene. If unset, "
        "the volume textures defining component weights will be assigned "
        "defaults likely unsuitable for atmosphere construction.",
        type=".SceneGeometry or None",
        init_type=".SceneGeometry or dict or str, optional",
        default="None",
    )

    def _eval_depolarization_factor_impl(self, si: SpectralIndex) -> np.ndarray:
        if isinstance(self.depolarization, t.Callable):
            depolarization = self.depolarization(si)
        elif isinstance(self.depolarization, np.ndarray):
            depolarization = np.atleast_1d(self.depolarization)
        elif self.depolarization is None:
            depolarization = np.zeros((1,))
        else:
            NotImplementedError

        return depolarization

    def eval_depolarization_factor(self, si: SpectralIndex) -> np.ndarray:
        """
        Evaluate the depolarization factor.

        Parameters
        ----------
        si: .SpectralIndex
            Spectral context.

        Returns
        -------
            Depolarization factor as an array of shape (N) where N can
            either be 1 or the number of layers of the atmosphere's vertical
            axis.
        """
        depolarization = self._eval_depolarization_factor_impl(si)
        return depolarization

    @property
    def template(self) -> dict:
        if eradiate.mode().is_polarized:
            result = {"type": "rayleigh_polarized"}

            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                if self.geometry is None:
                    to_world = mi.ScalarTransform4f()
                else:
                    to_world = self.geometry.atmosphere_volume_to_world

                result["depolarization.type"] = "gridvolume"
                result["depolarization.grid"] = DictParameter(
                    lambda ctx: mi.VolumeGrid(
                        np.reshape(
                            self.eval_depolarization_factor(ctx.si),
                            (-1, 1, 1),
                        ).astype(np.float32),
                    ),
                )
                result["depolarization.to_world"] = to_world
                result["depolarization.filter_type"] = "nearest"

            elif isinstance(self.geometry, SphericalShellGeometry):
                volume_rmin = self.geometry.atmosphere_volume_rmin
                to_world = self.geometry.atmosphere_volume_to_world

                result["depolarization.type"] = "sphericalcoordsvolume"
                result["depolarization.volume.type"] = "gridvolume"
                result["depolarization.volume.grid"] = DictParameter(
                    lambda ctx: mi.VolumeGrid(
                        np.reshape(
                            self.eval_depolarization_factor(ctx.si),
                            (-1, 1, 1),
                        ).astype(np.float32),
                    ),
                )
                result["depolarization.volume.filter_type"] = "nearest"
                result["depolarization.to_world"] = to_world
                result["depolarization.rmin"] = volume_rmin

            return result
        else:
            return {"type": "rayleigh"}

    @property
    def params(self) -> dict[str, SceneParameter]:
        result = {}

        if eradiate.mode().is_polarized:
            if self.geometry is None or isinstance(
                self.geometry, PlaneParallelGeometry
            ):
                result["depolarization.data"] = SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_depolarization_factor(ctx.si),
                        (-1, 1, 1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    # search=SearchSceneParameter(
                    #     node_type=mi.PhaseFunction,
                    #     node_id=self.phase.id,
                    #     parameter_relpath=,
                    # ),
                )

            elif isinstance(self.geometry, SphericalShellGeometry):
                result["depolarization.volume.data"] = SceneParameter(
                    lambda ctx: np.reshape(
                        self.eval_depolarization_factor(ctx.si),
                        (1, 1, -1, 1),
                    ).astype(np.float32),
                    KernelSceneParameterFlags.SPECTRAL,
                    # search=SearchSceneParameter(
                    #     node_type=mi.PhaseFunction,
                    #     node_id=self.phase.id,
                    #     parameter_relpath=,
                    # ),
                )
        return result
